from __future__ import annotations

import copy
import ctypes
import os
import platform
from _ctypes import _Pointer
from ctypes import (
    HRESULT,
    POINTER,
    Structure,
    byref,
    c_uint32,
    c_void_p,
    c_wchar_p,
    cast,
    create_unicode_buffer,
    pointer,
    sizeof,
    windll,
)
from ctypes.wintypes import BOOL, FLOAT, HDC, UINT, WCHAR
from typing import NoReturn

from pyglet.font.dwrite.d2d1_types_lib import (
    D2D1_COLOR_F,
    D2D1_POINT_2L,
    D2D1_SIZE_U,
    D2D_POINT_2F,
)
from pyglet.libs.win32 import INT16, INT32, LOGFONTW, UINT8, UINT16, UINT32, UINT64, c_void, com

try:
    dwrite = "dwrite"

    # System32 and SysWOW64 folders are opposite perception in Windows x64.
    # System32 = x64 dll's | SysWOW64 = x86 dlls
    # By default ctypes only seems to look in system32 regardless of Python architecture, which has x64 dlls.
    if platform.architecture()[0] == "32bit" and platform.machine().endswith("64"):  # Machine is x64, Python is x86.
        dwrite = os.path.join(os.environ["WINDIR"], "SysWOW64", "dwrite.dll")

    dwrite_lib = windll.LoadLibrary(dwrite)
except OSError:
    # Doesn't exist? Should stop import of library.
    msg = "DirectWrite Not Found"
    raise ImportError(msg)  # noqa: B904


def DWRITE_MAKE_OPENTYPE_TAG(a: str, b: str, c: str, d: str) -> int:
    return ord(d) << 24 | ord(c) << 16 | ord(b) << 8 | ord(a)

DWRITE_NO_PALETTE_INDEX = 0xFFFF

DWRITE_FACTORY_TYPE = UINT
DWRITE_FACTORY_TYPE_SHARED = 0
DWRITE_FACTORY_TYPE_ISOLATED = 1

DWRITE_FONT_WEIGHT = UINT
DWRITE_FONT_WEIGHT_THIN = 100
DWRITE_FONT_WEIGHT_EXTRA_LIGHT = 200
DWRITE_FONT_WEIGHT_ULTRA_LIGHT = 200
DWRITE_FONT_WEIGHT_LIGHT = 300
DWRITE_FONT_WEIGHT_SEMI_LIGHT = 350
DWRITE_FONT_WEIGHT_NORMAL = 400
DWRITE_FONT_WEIGHT_REGULAR = 400
DWRITE_FONT_WEIGHT_MEDIUM = 500
DWRITE_FONT_WEIGHT_DEMI_BOLD = 600
DWRITE_FONT_WEIGHT_SEMI_BOLD = 600
DWRITE_FONT_WEIGHT_BOLD = 700
DWRITE_FONT_WEIGHT_EXTRA_BOLD = 800
DWRITE_FONT_WEIGHT_ULTRA_BOLD = 800
DWRITE_FONT_WEIGHT_BLACK = 900
DWRITE_FONT_WEIGHT_HEAVY = 900
DWRITE_FONT_WEIGHT_EXTRA_BLACK = 950


DWRITE_FONT_STRETCH = UINT
DWRITE_FONT_STRETCH_UNDEFINED = 0
DWRITE_FONT_STRETCH_ULTRA_CONDENSED = 1
DWRITE_FONT_STRETCH_EXTRA_CONDENSED = 2
DWRITE_FONT_STRETCH_CONDENSED = 3
DWRITE_FONT_STRETCH_SEMI_CONDENSED = 4
DWRITE_FONT_STRETCH_NORMAL = 5
DWRITE_FONT_STRETCH_MEDIUM = 5
DWRITE_FONT_STRETCH_SEMI_EXPANDED = 6
DWRITE_FONT_STRETCH_EXPANDED = 7
DWRITE_FONT_STRETCH_EXTRA_EXPANDED = 8

DWRITE_GLYPH_IMAGE_FORMATS = UINT
DWRITE_GLYPH_IMAGE_FORMATS_NONE = 0x00000000
DWRITE_GLYPH_IMAGE_FORMATS_TRUETYPE = 0x00000001
DWRITE_GLYPH_IMAGE_FORMATS_CFF = 0x00000002
DWRITE_GLYPH_IMAGE_FORMATS_COLR = 0x00000004
DWRITE_GLYPH_IMAGE_FORMATS_SVG = 0x00000008
DWRITE_GLYPH_IMAGE_FORMATS_PNG = 0x00000010
DWRITE_GLYPH_IMAGE_FORMATS_JPEG = 0x00000020
DWRITE_GLYPH_IMAGE_FORMATS_TIFF = 0x00000040
DWRITE_GLYPH_IMAGE_FORMATS_PREMULTIPLIED_B8G8R8A8 = 0x00000080
DWRITE_GLYPH_IMAGE_FORMATS_COLR_PAINT_TREE = 0x00000100

DWRITE_GLYPH_IMAGE_FORMATS_BITMAP = (DWRITE_GLYPH_IMAGE_FORMATS_PNG |
                                     DWRITE_GLYPH_IMAGE_FORMATS_JPEG |
                                     DWRITE_GLYPH_IMAGE_FORMATS_TIFF |
                                     DWRITE_GLYPH_IMAGE_FORMATS_PREMULTIPLIED_B8G8R8A8)

DWRITE_MEASURING_MODE = UINT
DWRITE_MEASURING_MODE_NATURAL = 0
DWRITE_MEASURING_MODE_GDI_CLASSIC = 1
DWRITE_MEASURING_MODE_GDI_NATURAL = 2

DWRITE_GLYPH_IMAGE_FORMATS_ALL = DWRITE_GLYPH_IMAGE_FORMATS_TRUETYPE | \
                                 DWRITE_GLYPH_IMAGE_FORMATS_CFF | \
                                 DWRITE_GLYPH_IMAGE_FORMATS_COLR | \
                                 DWRITE_GLYPH_IMAGE_FORMATS_SVG | \
                                 DWRITE_GLYPH_IMAGE_FORMATS_PNG | \
                                 DWRITE_GLYPH_IMAGE_FORMATS_JPEG | \
                                 DWRITE_GLYPH_IMAGE_FORMATS_TIFF | \
                                 DWRITE_GLYPH_IMAGE_FORMATS_PREMULTIPLIED_B8G8R8A8

DWRITE_FONT_STYLE = UINT
DWRITE_FONT_STYLE_NORMAL = 0
DWRITE_FONT_STYLE_OBLIQUE = 1
DWRITE_FONT_STYLE_ITALIC = 2


DWRITE_INFORMATIONAL_STRING_ID = UINT32
DWRITE_INFORMATIONAL_STRING_NONE = 0
DWRITE_INFORMATIONAL_STRING_COPYRIGHT_NOTICE = 1
DWRITE_INFORMATIONAL_STRING_VERSION_STRINGS = 2
DWRITE_INFORMATIONAL_STRING_TRADEMARK = 3
DWRITE_INFORMATIONAL_STRING_MANUFACTURER = 4
DWRITE_INFORMATIONAL_STRING_DESIGNER = 5
DWRITE_INFORMATIONAL_STRING_DESIGNER_URL = 6
DWRITE_INFORMATIONAL_STRING_DESCRIPTION = 7
DWRITE_INFORMATIONAL_STRING_FONT_VENDOR_URL = 8
DWRITE_INFORMATIONAL_STRING_LICENSE_DESCRIPTION = 9
DWRITE_INFORMATIONAL_STRING_LICENSE_INFO_URL = 10
DWRITE_INFORMATIONAL_STRING_WIN32_FAMILY_NAMES = 11
DWRITE_INFORMATIONAL_STRING_WIN32_SUBFAMILY_NAMES = 12
DWRITE_INFORMATIONAL_STRING_TYPOGRAPHIC_FAMILY_NAMES = 13
DWRITE_INFORMATIONAL_STRING_TYPOGRAPHIC_SUBFAMILY_NAMES = 14
DWRITE_INFORMATIONAL_STRING_SAMPLE_TEXT = 15
DWRITE_INFORMATIONAL_STRING_FULL_NAME = 16
DWRITE_INFORMATIONAL_STRING_POSTSCRIPT_NAME = 17
DWRITE_INFORMATIONAL_STRING_POSTSCRIPT_CID_NAME = 18
DWRITE_INFORMATIONAL_STRING_WEIGHT_STRETCH_STYLE_FAMILY_NAME = 19
DWRITE_INFORMATIONAL_STRING_DESIGN_SCRIPT_LANGUAGE_TAG = 20
DWRITE_INFORMATIONAL_STRING_SUPPORTED_SCRIPT_LANGUAGE_TAG = 21
DWRITE_INFORMATIONAL_STRING_PREFERRED_FAMILY_NAMES = 22
DWRITE_INFORMATIONAL_STRING_PREFERRED_SUBFAMILY_NAMES = 23
DWRITE_INFORMATIONAL_STRING_WWS_FAMILY_NAME = 24


DWRITE_TEXT_ALIGNMENT = UINT
DWRITE_TEXT_ALIGNMENT_LEADING = 1
DWRITE_TEXT_ALIGNMENT_TRAILING = 2
DWRITE_TEXT_ALIGNMENT_CENTER = 3
DWRITE_TEXT_ALIGNMENT_JUSTIFIED = 4

DWRITE_FONT_FACE_TYPE = UINT
DWRITE_FONT_FACE_TYPE_CFF = 0
DWRITE_FONT_FACE_TYPE_TRUETYPE = 1
DWRITE_FONT_FACE_TYPE_OPENTYPE_COLLECTION = 2
DWRITE_FONT_FACE_TYPE_TYPE1 = 3
DWRITE_FONT_FACE_TYPE_VECTOR = 4
DWRITE_FONT_FACE_TYPE_BITMAP = 5
DWRITE_FONT_FACE_TYPE_UNKNOWN = 6
DWRITE_FONT_FACE_TYPE_RAW_CFF = 7
DWRITE_FONT_FACE_TYPE_TRUETYPE_COLLECTION = 8

def repr_func(self):
    field_values = []
    for field_value in self._fields_:
        try:
            name, typ, _ = field_value  # bitfields
        except ValueError:
            name, typ = field_value

        value = getattr(self, name)

        # Check if the field is a POINTER to any ctypes structure
        if isinstance(value, _Pointer):
            value = repr(value.contents) if value else "NULL"

        field_values.append(f"{name}={value}")

    return f"{self.__class__.__name__}({', '.join(field_values)})"

DWRITE_COLOR_F = D2D1_COLOR_F

class DWRITE_TEXT_METRICS(Structure):
    _fields_ = (
        ("left", FLOAT),
        ("top", FLOAT),
        ("width", FLOAT),
        ("widthIncludingTrailingWhitespace", FLOAT),
        ("height", FLOAT),
        ("layoutWidth", FLOAT),
        ("layoutHeight", FLOAT),
        ("maxBidiReorderingDepth", UINT32),
        ("lineCount", UINT32),
    )


class DWRITE_FONT_METRICS(Structure):
    _fields_ = (
        ("designUnitsPerEm", UINT16),
        ("ascent", UINT16),
        ("descent", UINT16),
        ("lineGap", INT16),
        ("capHeight", UINT16),
        ("xHeight", UINT16),
        ("underlinePosition", INT16),
        ("underlineThickness", UINT16),
        ("strikethroughPosition", INT16),
        ("strikethroughThickness", UINT16),
    )

class DWRITE_FONT_METRICS1(Structure):
    _fields_ = (
        ("designUnitsPerEm", UINT16),
        ("ascent", UINT16),
        ("descent", UINT16),
        ("lineGap", INT16),
        ("capHeight", UINT16),
        ("xHeight", UINT16),
        ("underlinePosition", INT16),
        ("underlineThickness", UINT16),
        ("strikethroughPosition", INT16),
        ("strikethroughThickness", UINT16),
        ("glyphBoxLeft", INT16),  # New starting here.
        ("glyphBoxTop", INT16),
        ("glyphBoxRight", INT16),
        ("glyphBoxBottom", INT16),
        ("subscriptPositionX", INT16),
        ("subscriptPositionY", INT16),
        ("subscriptSizeX", INT16),
        ("subscriptSizeY", INT16),
        ("superscriptPositionX", INT16),
        ("superscriptPositionY", INT16),
        ("superscriptSizeX", INT16),
        ("superscriptSizeY", INT16),
        ("hasTypographicMetrics", BOOL),
    )

class DWRITE_CARET_METRICS(Structure):
    _fields_ = (
        ("slopeRise", INT16),
        ("slopeRun", INT16),
        ("offset", INT16),
    )

class DWRITE_GLYPH_METRICS(Structure):
    _fields_ = (
        ("leftSideBearing", INT32),
        ("advanceWidth", UINT32),
        ("rightSideBearing", INT32),
        ("topSideBearing", INT32),
        ("advanceHeight", UINT32),
        ("bottomSideBearing", INT32),
        ("verticalOriginY", INT32),
    )

    def __repr__(self):
        return (f"DWRITE_GLYPH_METRICS(leftSideBearing={self.leftSideBearing}, advanceWidth={self.advanceWidth}, "
                f"rightSideBearing={self.rightSideBearing}, topSideBearing={self.topSideBearing}, advanceHeight={self.advanceHeight}, "
                f"bottomSideBearing={self.bottomSideBearing}, verticalOriginY={self.verticalOriginY})")


class DWRITE_GLYPH_OFFSET(Structure):
    _fields_ = (
        ("advanceOffset", FLOAT),
        ("ascenderOffset", FLOAT),
    )

    def __repr__(self) -> str:
        return f"DWRITE_GLYPH_OFFSET({self.advanceOffset}, {self.ascenderOffset})"


class DWRITE_CLUSTER_METRICS(Structure):
    _fields_ = (
        ("width", FLOAT),
        ("length", UINT16),
        ("canWrapLineAfter", UINT16, 1),
        ("isWhitespace", UINT16, 1),
        ("isNewline", UINT16, 1),
        ("isSoftHyphen", UINT16, 1),
        ("isRightToLeft", UINT16, 1),
        ("padding", UINT16, 11),
    )

DWRITE_CLUSTER_METRICS.__repr__ = repr_func

class IDWriteFontFileStream(com.IUnknown):
    _methods_ = [
        ("ReadFileFragment",
         com.STDMETHOD(POINTER(c_void_p), UINT64, UINT64, POINTER(c_void_p))),
        ("ReleaseFileFragment",
         com.STDMETHOD(c_void_p)),
        ("GetFileSize",
         com.STDMETHOD(POINTER(UINT64))),
        ("GetLastWriteTime",
         com.STDMETHOD(POINTER(UINT64))),
    ]


class IDWriteFontFileLoader_LI(com.IUnknown):  # Local implementation use only.
    _methods_ = [
        ("CreateStreamFromKey",
         com.STDMETHOD(c_void_p, UINT32, POINTER(POINTER(IDWriteFontFileStream)))),
    ]


class IDWriteFontFileLoader(com.pIUnknown):
    _methods_ = [
        ("CreateStreamFromKey",
         com.STDMETHOD(c_void_p, UINT32, POINTER(POINTER(IDWriteFontFileStream)))),
    ]


class IDWriteLocalFontFileLoader(IDWriteFontFileLoader, com.pIUnknown):
    _methods_ = [
        ("GetFilePathLengthFromKey",
         com.STDMETHOD(c_void_p, UINT32, POINTER(UINT32))),
        ("GetFilePathFromKey",
         com.STDMETHOD(c_void_p, UINT32, c_wchar_p, UINT32)),
        ("GetLastWriteTimeFromKey",
         com.STDMETHOD()),
    ]


IID_IDWriteLocalFontFileLoader = com.GUID(0xb2d9f3ec, 0xc9fe, 0x4a11, 0xa2, 0xec, 0xd8, 0x62, 0x08, 0xf7, 0xc0, 0xa2)


class IDWriteLocalizedStrings(com.pIUnknown):
    _methods_ = [
        ("GetCount",
         com.METHOD(UINT32)),
        ("FindLocaleName",
         com.STDMETHOD(c_wchar_p, POINTER(UINT32), POINTER(BOOL))),
        ("GetLocaleNameLength",
         com.STDMETHOD(UINT32, POINTER(UINT32))),
        ("GetLocaleName",
         com.STDMETHOD(UINT32, c_wchar_p, UINT32)),
        ("GetStringLength",
         com.STDMETHOD(UINT32, POINTER(UINT32))),
        ("GetString",
         com.STDMETHOD(UINT32, c_wchar_p, UINT32)),
    ]

class IDWriteFontFile(com.pIUnknown):
    _methods_ = [
        ("GetReferenceKey",
         com.STDMETHOD(POINTER(c_void_p), POINTER(UINT32))),
        ("GetLoader",
         com.STDMETHOD(POINTER(IDWriteFontFileLoader))),
        ("Analyze",
         com.STDMETHOD()),
    ]


class IDWriteFontFace(com.pIUnknown):
    _methods_ = [
        ("GetType",
         com.METHOD(DWRITE_FONT_FACE_TYPE)),
        ("GetFiles",
         com.STDMETHOD(POINTER(UINT32), POINTER(IDWriteFontFile))),
        ("GetIndex",
         com.METHOD(UINT32)),
        ("GetSimulations",
         com.STDMETHOD()),
        ("IsSymbolFont",
         com.METHOD(BOOL)),
        ("GetMetrics",
         com.METHOD(c_void, POINTER(DWRITE_FONT_METRICS))),
        ("GetGlyphCount",
         com.METHOD(UINT16)),
        ("GetDesignGlyphMetrics",
         com.STDMETHOD(POINTER(UINT16), UINT32, POINTER(DWRITE_GLYPH_METRICS), BOOL)),
        ("GetGlyphIndices",
         com.STDMETHOD(POINTER(UINT32), UINT32, POINTER(UINT16))),
        ("TryGetFontTable",
         com.STDMETHOD(UINT32, c_void_p, POINTER(UINT32), c_void_p, POINTER(BOOL))),
        ("ReleaseFontTable",
         com.METHOD(c_void)),
        ("GetGlyphRunOutline",
         com.STDMETHOD()),
        ("GetRecommendedRenderingMode",
         com.STDMETHOD()),
        ("GetGdiCompatibleMetrics",
         com.STDMETHOD()),
        ("GetGdiCompatibleGlyphMetrics",
         com.STDMETHOD()),
    ]


IID_IDWriteFontFace1 = com.GUID(0xa71efdb4, 0x9fdb, 0x4838, 0xad, 0x90, 0xcf, 0xc3, 0xbe, 0x8c, 0x3d, 0xaf)


class IDWriteFontFace1(IDWriteFontFace):
    _methods_ = [
        ("GetMetrics1",
         com.STDMETHOD(POINTER(DWRITE_FONT_METRICS1))),
        ("GetGdiCompatibleMetrics1",
         com.STDMETHOD()),
        ("GetCaretMetrics",
         com.STDMETHOD(POINTER(DWRITE_CARET_METRICS))),
        ("GetUnicodeRanges",
         com.STDMETHOD()),
        ("IsMonospacedFont",
         com.METHOD(BOOL)),
        ("GetDesignGlyphAdvances",
         com.METHOD(c_void, POINTER(DWRITE_FONT_METRICS))),
        ("GetGdiCompatibleGlyphAdvances",
         com.STDMETHOD()),
        ("GetKerningPairAdjustments",
         com.STDMETHOD(UINT32, POINTER(UINT16), POINTER(INT32))),
        ("HasKerningPairs",
         com.METHOD(BOOL)),
        ("GetRecommendedRenderingMode1",
         com.STDMETHOD()),
        ("GetVerticalGlyphVariants",
         com.STDMETHOD()),
        ("HasVerticalGlyphVariants",
         com.METHOD(BOOL)),
    ]

IID_IDWriteFontFace2 = com.GUID(0xd8b768ff, 0x64bc, 0x4e66, 0x98,0x2b, 0xec,0x8e,0x87,0xf6,0x93,0xf7)

class IDWriteFontFace2(IDWriteFontFace1):
    _methods_ = [
        ("IsColorFont",
         com.METHOD(BOOL)),
        ("GetColorPaletteCount",
         com.METHOD(UINT32)),
        ("GetPaletteEntryCount",
         com.METHOD(UINT32)),
        ("GetPaletteEntries",
         com.STDMETHOD()),
        ("GetRecommendedRenderingMode2",
         com.STDMETHOD()),
    ]

IID_IDWriteFontFace3 = com.GUID(0xd37d7598, 0x09be, 0x4222, 0xa2,0x36, 0x20,0x81,0x34,0x1c,0xc1,0xf2)

class IDWriteFontFace3(IDWriteFontFace2):
    _methods_ = [
        ("GetFontFaceReference",
         com.STDMETHOD()),
        ("GetPanose",
         com.STDMETHOD()),
        ("GetWeight",
         com.METHOD(DWRITE_FONT_WEIGHT)),
        ("GetStretch",
         com.METHOD(DWRITE_FONT_STRETCH)),
        ("GetStyle",
         com.METHOD(DWRITE_FONT_STYLE)),
        ("GetFamilyNames",
         com.STDMETHOD(POINTER(IDWriteLocalizedStrings))),
        ("GetFaceNames",
         com.STDMETHOD(POINTER(IDWriteLocalizedStrings))),
        ("GetInformationalStrings",
         com.STDMETHOD()),
        ("HasCharacter",
         com.METHOD(BOOL, UINT32)),
        ("GetRecommendedRenderingMode3",
         com.STDMETHOD()),
        ("IsCharacterLocal",
         com.STDMETHOD()),
        ("IsGlyphLocal",
         com.STDMETHOD()),
        ("AreCharactersLocal",
         com.STDMETHOD()),
        ("AreGlyphsLocal",
         com.STDMETHOD()),
    ]


class DWRITE_GLYPH_IMAGE_DATA(Structure):
    _fields_ = [
        ("imageData", c_void_p),
        ("imageDataSize", UINT32),
        ("uniqueDataId", UINT32),
        ("pixelsPerEm", UINT32),
        ("pixelSize", D2D1_SIZE_U),
        ("horizontalLeftOrigin", D2D1_POINT_2L),
        ("horizontalRightOrigin", D2D1_POINT_2L),
        ("verticalTopOrigin", D2D1_POINT_2L),
        ("verticalBottomOrigin", D2D1_POINT_2L),
    ]

IID_IDWriteFontFace4 = com.GUID(0x27f2a904, 0x4eb8, 0x441d, 0x96,0x78, 0x05,0x63,0xf5,0x3e,0x3e,0x2f)

class IDWriteFontFace4(IDWriteFontFace3):
    _methods_ = [
        ("GetGlyphImageFormats_",
         com.STDMETHOD(UINT16, UINT32, UINT32, POINTER(DWRITE_GLYPH_IMAGE_FORMATS))),
        ("GetGlyphImageFormats",
         com.METHOD(DWRITE_GLYPH_IMAGE_FORMATS)),
        ("GetGlyphImageData",
         com.STDMETHOD(UINT16, UINT32, DWRITE_GLYPH_IMAGE_FORMATS, POINTER(DWRITE_GLYPH_IMAGE_DATA),
                       POINTER(c_void_p))),
        ("ReleaseGlyphImageData",
         com.METHOD(None, c_void_p))
    ]


class DWRITE_GLYPH_RUN(Structure):
    _fields_ = (
        ("fontFace", IDWriteFontFace),
        ("fontEmSize", FLOAT),
        ("glyphCount", UINT32),
        ("glyphIndices", POINTER(UINT16)),
        ("glyphAdvances", POINTER(FLOAT)),
        ("glyphOffsets", POINTER(DWRITE_GLYPH_OFFSET)),
        ("isSideways", BOOL),
        ("bidiLevel", UINT32),
    )

    def get_info(self):
        print(f"DWRITE_GLYPH_RUN(face: {self.fontFace}, fontEmSize: {self.fontEmSize}, count: {self.glyphCount}")
        for i in range(self.glyphCount):
            print(f"glyph: {i}, indice: {self.glyphIndices[i]}, advance: {self.glyphAdvances[i]}, offset: {self.glyphOffsets[i]}")


DWRITE_SCRIPT_SHAPES = UINT
DWRITE_SCRIPT_SHAPES_DEFAULT = 0


class DWRITE_SCRIPT_ANALYSIS(Structure):
    _fields_ = (
        ("script", UINT16),
        ("shapes", DWRITE_SCRIPT_SHAPES),
    )


DWRITE_FONT_FEATURE_TAG = UINT


class DWRITE_FONT_FEATURE(Structure):
    _fields_ = (
        ("nameTag", DWRITE_FONT_FEATURE_TAG),
        ("parameter", UINT32),
    )


class DWRITE_TYPOGRAPHIC_FEATURES(Structure):
    _fields_ = (
        ("features", POINTER(DWRITE_FONT_FEATURE)),
        ("featureCount", UINT32),
    )


class DWRITE_SHAPING_TEXT_PROPERTIES(Structure):
    _fields_ = (
        ("isShapedAlone", UINT16, 1),
        ("reserved1", UINT16, 1),
        ("canBreakShapingAfter", UINT16, 1),
        ("reserved", UINT16, 13),
    )

    def __repr__(self) -> str:
        return f"DWRITE_SHAPING_TEXT_PROPERTIES({self.isShapedAlone}, {self.reserved1}, {self.canBreakShapingAfter})"


class DWRITE_SHAPING_GLYPH_PROPERTIES(Structure):
    _fields_ = (
        ("justification", UINT16, 4),
        ("isClusterStart", UINT16, 1),
        ("isDiacritic", UINT16, 1),
        ("isZeroWidthSpace", UINT16, 1),
        ("reserved", UINT16, 9),
    )
DWRITE_SHAPING_GLYPH_PROPERTIES.__repr__ = repr_func

DWRITE_READING_DIRECTION = UINT
DWRITE_READING_DIRECTION_LEFT_TO_RIGHT = 0


class IDWriteTextAnalysisSource(com.IUnknown):
    _methods_ = [
        ("GetTextAtPosition",
         com.STDMETHOD(UINT32, POINTER(c_wchar_p), POINTER(UINT32))),
        ("GetTextBeforePosition",
         com.STDMETHOD(UINT32, POINTER(c_wchar_p), POINTER(UINT32))),
        ("GetParagraphReadingDirection",
         com.METHOD(DWRITE_READING_DIRECTION)),
        ("GetLocaleName",
         com.STDMETHOD(UINT32, POINTER(UINT32), POINTER(c_wchar_p))),
        ("GetNumberSubstitution",
         com.STDMETHOD(UINT32, POINTER(UINT32), c_void_p)),
    ]


class IDWriteTextAnalysisSink(com.IUnknown):
    _methods_ = [
        ("SetScriptAnalysis",
         com.STDMETHOD(UINT32, UINT32, POINTER(DWRITE_SCRIPT_ANALYSIS))),
        ("SetLineBreakpoints",
         com.STDMETHOD(UINT32, UINT32, c_void_p)),
        ("SetBidiLevel",
         com.STDMETHOD(UINT32, UINT32, UINT8, UINT8)),
        ("SetNumberSubstitution",
         com.STDMETHOD(UINT32, UINT32, c_void_p)),
    ]


class IDWriteTextAnalyzer(com.pIUnknown):
    _methods_ = [
        ("AnalyzeScript",
         com.STDMETHOD(POINTER(IDWriteTextAnalysisSource), UINT32, UINT32, POINTER(IDWriteTextAnalysisSink))),
        ("AnalyzeBidi",
         com.STDMETHOD()),
        ("AnalyzeNumberSubstitution",
         com.STDMETHOD()),
        ("AnalyzeLineBreakpoints",
         com.STDMETHOD()),
        ("GetGlyphs",
         com.STDMETHOD(c_wchar_p, UINT32, IDWriteFontFace, BOOL, BOOL, POINTER(DWRITE_SCRIPT_ANALYSIS),
                       c_wchar_p, c_void_p, POINTER(POINTER(DWRITE_TYPOGRAPHIC_FEATURES)), POINTER(UINT32),
                       UINT32, UINT32, POINTER(UINT16), POINTER(DWRITE_SHAPING_TEXT_PROPERTIES),
                       POINTER(UINT16), POINTER(DWRITE_SHAPING_GLYPH_PROPERTIES), POINTER(UINT32))),
        ("GetGlyphPlacements",
         com.STDMETHOD(c_wchar_p, POINTER(UINT16), POINTER(DWRITE_SHAPING_TEXT_PROPERTIES), UINT32, POINTER(UINT16),
                       POINTER(DWRITE_SHAPING_GLYPH_PROPERTIES), UINT32, IDWriteFontFace, FLOAT, BOOL, BOOL,
                       POINTER(DWRITE_SCRIPT_ANALYSIS), c_wchar_p, POINTER(DWRITE_TYPOGRAPHIC_FEATURES),
                       POINTER(UINT32), UINT32, POINTER(FLOAT), POINTER(DWRITE_GLYPH_OFFSET))),
        ("GetGdiCompatibleGlyphPlacements",
         com.STDMETHOD()),
    ]




class IDWriteFontList(com.pIUnknown):
    _methods_ = [
        ("GetFontCollection",
         com.STDMETHOD()),
        ("GetFontCount",
         com.METHOD(UINT32)),
        ("GetFont",
         com.STDMETHOD(UINT32, c_void_p)),  # IDWriteFont, use void because of forward ref.
    ]


class IDWriteFontFamily(IDWriteFontList, com.pIUnknown):
    _methods_ = [
        ("GetFamilyNames",
         com.STDMETHOD(POINTER(IDWriteLocalizedStrings))),
        ("GetFirstMatchingFont",
         com.STDMETHOD(DWRITE_FONT_WEIGHT, DWRITE_FONT_STRETCH, DWRITE_FONT_STYLE, c_void_p)),
        ("GetMatchingFonts",
         com.STDMETHOD()),
    ]


class IDWriteFontFamily1(IDWriteFontFamily, IDWriteFontList, com.pIUnknown):
    _methods_ = [
        ("GetFontLocality",
         com.STDMETHOD()),
        ("GetFont1",
         com.STDMETHOD()),
        ("GetFontFaceReference",
         com.STDMETHOD()),
    ]


class IDWriteFont(com.pIUnknown):
    _methods_ = [
        ("GetFontFamily",
         com.STDMETHOD(POINTER(IDWriteFontFamily))),
        ("GetWeight",
         com.METHOD(DWRITE_FONT_WEIGHT)),
        ("GetStretch",
         com.METHOD(DWRITE_FONT_STRETCH)),
        ("GetStyle",
         com.METHOD(DWRITE_FONT_STYLE)),
        ("IsSymbolFont",
         com.METHOD(BOOL)),
        ("GetFaceNames",
         com.STDMETHOD(POINTER(IDWriteLocalizedStrings))),
        ("GetInformationalStrings",
         com.STDMETHOD(DWRITE_INFORMATIONAL_STRING_ID, POINTER(IDWriteLocalizedStrings), POINTER(BOOL))),
        ("GetSimulations",
         com.STDMETHOD()),
        ("GetMetrics",
         com.STDMETHOD()),
        ("HasCharacter",
         com.STDMETHOD(UINT32, POINTER(BOOL))),
        ("CreateFontFace",
         com.STDMETHOD(POINTER(IDWriteFontFace))),
    ]


class IDWriteFont1(IDWriteFont, com.pIUnknown):
    _methods_ = [
        ("GetMetrics1",
         com.STDMETHOD()),
        ("GetPanose",
         com.STDMETHOD()),
        ("GetUnicodeRanges",
         com.STDMETHOD()),
        ("IsMonospacedFont",
         com.STDMETHOD()),
    ]


class IDWriteFontCollection(com.pIUnknown):
    _methods_ = [
        ("GetFontFamilyCount",
         com.METHOD(UINT32)),
        ("GetFontFamily",
         com.STDMETHOD(UINT32, POINTER(IDWriteFontFamily))),
        ("FindFamilyName",
         com.STDMETHOD(c_wchar_p, POINTER(UINT), POINTER(BOOL))),
        ("GetFontFromFontFace",
         com.STDMETHOD()),
    ]


class IDWriteFontCollection1(IDWriteFontCollection, com.pIUnknown):
    _methods_ = [
        ("GetFontSet",
         com.STDMETHOD()),
        ("GetFontFamily1",
         com.STDMETHOD(POINTER(IDWriteFontFamily1))),
    ]


class IDWriteGdiInterop(com.pIUnknown):
    _methods_ = [
        ("CreateFontFromLOGFONT",
         com.STDMETHOD(POINTER(LOGFONTW), POINTER(IDWriteFont))),
        ("ConvertFontToLOGFONT",
         com.STDMETHOD()),
        ("ConvertFontFaceToLOGFONT",
         com.STDMETHOD()),
        ("CreateFontFaceFromHdc",
         com.STDMETHOD(HDC, POINTER(IDWriteFontFace))),
        ("CreateBitmapRenderTarget",
         com.STDMETHOD()),
    ]


class IDWriteTextFormat(com.pIUnknown):
    _methods_ = [
        ("SetTextAlignment",
         com.STDMETHOD(DWRITE_TEXT_ALIGNMENT)),
        ("SetParagraphAlignment",
         com.STDMETHOD()),
        ("SetWordWrapping",
         com.STDMETHOD()),
        ("SetReadingDirection",
         com.STDMETHOD()),
        ("SetFlowDirection",
         com.STDMETHOD()),
        ("SetIncrementalTabStop",
         com.STDMETHOD()),
        ("SetTrimming",
         com.STDMETHOD()),
        ("SetLineSpacing",
         com.STDMETHOD()),
        ("GetTextAlignment",
         com.STDMETHOD()),
        ("GetParagraphAlignment",
         com.STDMETHOD()),
        ("GetWordWrapping",
         com.STDMETHOD()),
        ("GetReadingDirection",
         com.STDMETHOD()),
        ("GetFlowDirection",
         com.STDMETHOD()),
        ("GetIncrementalTabStop",
         com.STDMETHOD()),
        ("GetTrimming",
         com.STDMETHOD()),
        ("GetLineSpacing",
         com.STDMETHOD()),
        ("GetFontCollection",
         com.STDMETHOD()),
        ("GetFontFamilyNameLength",
         com.STDMETHOD(UINT32, POINTER(UINT32))),
        ("GetFontFamilyName",
         com.STDMETHOD(UINT32, c_wchar_p, UINT32)),
        ("GetFontWeight",
         com.STDMETHOD()),
        ("GetFontStyle",
         com.STDMETHOD()),
        ("GetFontStretch",
         com.STDMETHOD()),
        ("GetFontSize",
         com.STDMETHOD()),
        ("GetLocaleNameLength",
         com.STDMETHOD()),
        ("GetLocaleName",
         com.STDMETHOD()),
    ]


class IDWriteTypography(com.pIUnknown):
    _methods_ = [
        ("AddFontFeature",
         com.STDMETHOD(DWRITE_FONT_FEATURE)),
        ("GetFontFeatureCount",
         com.METHOD(UINT32)),
        ("GetFontFeature",
         com.STDMETHOD()),
    ]


class DWRITE_TEXT_RANGE(Structure):
    _fields_ = (
        ("startPosition", UINT32),
        ("length", UINT32),
    )


class DWRITE_OVERHANG_METRICS(Structure):
    _fields_ = (
        ("left", FLOAT),
        ("top", FLOAT),
        ("right", FLOAT),
        ("bottom", FLOAT),
    )


class DWRITE_GLYPH_RUN_DESCRIPTION(Structure):
    _fields_ = [
        ('localeName', c_wchar_p),
        ('text', c_wchar_p),
        ('textLength', UINT32),
        ('clusterMap', POINTER(UINT16)),
        ('textPosition', UINT32)
    ]

class DWRITE_COLOR_GLYPH_RUN(Structure):
    _fields_ = [
        ('glyphRun', DWRITE_GLYPH_RUN),
        ('glyphRunDescription', POINTER(DWRITE_GLYPH_RUN_DESCRIPTION)),
        ('baselineOriginX', FLOAT),
        ('baselineOriginY', FLOAT),
        ('runColor', DWRITE_COLOR_F),
        ('paletteIndex', UINT16),
    ]


DWRITE_COLOR_GLYPH_RUN1_Fields = [
    ('glyphRun', DWRITE_GLYPH_RUN),
    ('glyphRunDescription', POINTER(DWRITE_GLYPH_RUN_DESCRIPTION)),
    ('baselineOriginX', FLOAT),
    ('baselineOriginY', FLOAT),
    ('runColor', DWRITE_COLOR_F),
    ('paletteIndex', UINT16),
    ("_pad", UINT32),  # Padding only exists in 64 bit... only info found in header.
    ('glyphImageFormat', DWRITE_GLYPH_IMAGE_FORMATS),
    ('measuringMode', DWRITE_MEASURING_MODE)
]

if platform.architecture()[0] == "32bit":
    DWRITE_COLOR_GLYPH_RUN1_Fields.remove(("_pad", UINT32))


class DWRITE_COLOR_GLYPH_RUN1(Structure):
    _fields_ = DWRITE_COLOR_GLYPH_RUN1_Fields

class DWRITE_MATRIX(Structure):
    _fields_ = [
        ("m11", FLOAT),
        ("m12", FLOAT),
        ("m21", FLOAT),
        ("m22", FLOAT),
        ("dx", FLOAT),
        ("dy", FLOAT),
    ]

class IDWritePixelSnapping(com.IUnknown):
    _methods_ = [
        ("IsPixelSnappingDisabled", com.STDMETHOD(c_void_p, POINTER(FLOAT))),
        ("GetCurrentTransform", com.STDMETHOD(c_void_p, POINTER(DWRITE_MATRIX))),
        ("GetPixelsPerDip", com.STDMETHOD(c_void_p, POINTER(FLOAT))),
    ]

class IDWriteTextRenderer(IDWritePixelSnapping):
    _methods_ = [
        ("DrawGlyphRun",
            com.STDMETHOD(c_void_p, FLOAT, FLOAT, DWRITE_MEASURING_MODE,
                          POINTER(DWRITE_GLYPH_RUN),
                          POINTER(DWRITE_GLYPH_RUN_DESCRIPTION), c_void_p)),
        ("DrawUnderline",
         com.STDMETHOD(c_void_p, FLOAT, FLOAT, c_void_p, c_void_p)
         ),
        ("DrawStrikethrough",
         com.STDMETHOD(c_void_p, FLOAT, FLOAT, c_void_p, c_void_p)),
        ("DrawInlineObject",
         com.STDMETHOD(c_void_p, FLOAT, FLOAT, c_void_p, BOOL, BOOL, c_void_p)),
    ]


class IDWriteTextLayout(IDWriteTextFormat):
    _methods_ = [
        ("SetMaxWidth",
         com.STDMETHOD()),
        ("SetMaxHeight",
         com.STDMETHOD()),
        ("SetFontCollection",
         com.STDMETHOD()),
        ("SetFontFamilyName",
         com.STDMETHOD()),
        ("SetFontWeight",  # 30
         com.STDMETHOD()),
        ("SetFontStyle",
         com.STDMETHOD()),
        ("SetFontStretch",
         com.STDMETHOD()),
        ("SetFontSize",
         com.STDMETHOD()),
        ("SetUnderline",
         com.STDMETHOD()),
        ("SetStrikethrough",
         com.STDMETHOD()),
        ("SetDrawingEffect",
         com.STDMETHOD()),
        ("SetInlineObject",
         com.STDMETHOD()),
        ("SetTypography",
         com.STDMETHOD(IDWriteTypography, DWRITE_TEXT_RANGE)),
        ("SetLocaleName",
         com.STDMETHOD()),
        ("GetMaxWidth",  # 40
         com.METHOD(FLOAT)),
        ("GetMaxHeight",
         com.METHOD(FLOAT)),
        ("GetFontCollection2",
         com.STDMETHOD(UINT32, POINTER(IDWriteFontCollection), POINTER(DWRITE_TEXT_RANGE))),
        ("GetFontFamilyNameLength2",
         com.STDMETHOD(UINT32, POINTER(UINT32), c_void_p)),
        ("GetFontFamilyName2",
         com.STDMETHOD(UINT32, c_wchar_p, UINT32, POINTER(DWRITE_TEXT_RANGE))),
        ("GetFontWeight2",
         com.STDMETHOD(UINT32, POINTER(DWRITE_FONT_WEIGHT), POINTER(DWRITE_TEXT_RANGE))),
        ("GetFontStyle2",
         com.STDMETHOD()),
        ("GetFontStretch2",
         com.STDMETHOD()),
        ("GetFontSize2",
         com.STDMETHOD()),
        ("GetUnderline",
         com.STDMETHOD()),
        ("GetStrikethrough",
         com.STDMETHOD(UINT32, POINTER(BOOL), POINTER(DWRITE_TEXT_RANGE))),
        ("GetDrawingEffect",
         com.STDMETHOD()),
        ("GetInlineObject",
         com.STDMETHOD()),
        ("GetTypography",  # Always returns NULL without SetTypography being called.
         com.STDMETHOD(UINT32, POINTER(IDWriteTypography), POINTER(DWRITE_TEXT_RANGE))),
        ("GetLocaleNameLength1",
         com.STDMETHOD()),
        ("GetLocaleName1",
         com.STDMETHOD()),
        ("Draw",
         com.STDMETHOD(c_void_p, POINTER(IDWriteTextRenderer), FLOAT, FLOAT)),
        ("GetLineMetrics",
         com.STDMETHOD()),
        ("GetMetrics",
         com.STDMETHOD(POINTER(DWRITE_TEXT_METRICS))),
        ("GetOverhangMetrics",
         com.STDMETHOD(POINTER(DWRITE_OVERHANG_METRICS))),
        ("GetClusterMetrics",
         com.STDMETHOD(POINTER(DWRITE_CLUSTER_METRICS), UINT32, POINTER(UINT32))),
        ("DetermineMinWidth",
         com.STDMETHOD(POINTER(FLOAT))),
        ("HitTestPoint",
         com.STDMETHOD()),
        ("HitTestTextPosition",
         com.STDMETHOD()),
        ("HitTestTextRange",
         com.STDMETHOD()),
    ]


class IDWriteTextLayout1(IDWriteTextLayout, IDWriteTextFormat, com.pIUnknown):
    _methods_ = [
        ("SetPairKerning",
         com.STDMETHOD()),
        ("GetPairKerning",
         com.STDMETHOD()),
        ("SetCharacterSpacing",
         com.STDMETHOD()),
        ("GetCharacterSpacing",
         com.STDMETHOD(UINT32, POINTER(FLOAT), POINTER(FLOAT), POINTER(FLOAT), POINTER(DWRITE_TEXT_RANGE))),
    ]


class IDWriteFontFileEnumerator(com.IUnknown):
    _methods_ = [
        ("MoveNext",
         com.STDMETHOD(POINTER(BOOL))),
        ("GetCurrentFontFile",
         com.STDMETHOD(c_void_p)),
    ]


class IDWriteFontCollectionLoader(com.IUnknown):
    _methods_ = [
        ("CreateEnumeratorFromKey",
         com.STDMETHOD(c_void_p, c_void_p, UINT32, POINTER(POINTER(IDWriteFontFileEnumerator)))),
    ]

# Windows 7
IID_IDWriteFactory = com.GUID(0xb859ee5a, 0xd838, 0x4b5b, 0xa2, 0xe8, 0x1a, 0xdc, 0x7d, 0x93, 0xdb, 0x48)


class IDWriteRenderingParams(com.pIUnknown):
    _methods_ = [
        ("GetGamma",
         com.METHOD(FLOAT)),
        ("GetEnhancedContrast",
         com.METHOD(FLOAT)),
        ("GetClearTypeLevel",
         com.METHOD(FLOAT)),
        ("GetPixelGeometry",
         com.METHOD(UINT)),
        ("GetRenderingMode",
         com.METHOD(UINT)),
    ]


class IDWriteFactory(com.pIUnknown):
    _methods_ = [
        ("GetSystemFontCollection",
         com.STDMETHOD(POINTER(IDWriteFontCollection), BOOL)),
        ("CreateCustomFontCollection",
         com.STDMETHOD(POINTER(IDWriteFontCollectionLoader), c_void_p, UINT32, POINTER(IDWriteFontCollection))),
        ("RegisterFontCollectionLoader",
         com.STDMETHOD(POINTER(IDWriteFontCollectionLoader))),
        ("UnregisterFontCollectionLoader",
         com.STDMETHOD(POINTER(IDWriteFontCollectionLoader))),
        ("CreateFontFileReference",
         com.STDMETHOD(c_wchar_p, c_void_p, POINTER(IDWriteFontFile))),
        ("CreateCustomFontFileReference",
         com.STDMETHOD(c_void_p, UINT32, POINTER(IDWriteFontFileLoader_LI), POINTER(IDWriteFontFile))),
        ("CreateFontFace",
         com.STDMETHOD()),
        ("CreateRenderingParams",
         com.STDMETHOD(POINTER(IDWriteRenderingParams))),
        ("CreateMonitorRenderingParams",
         com.STDMETHOD()),
        ("CreateCustomRenderingParams",
         com.STDMETHOD(FLOAT, FLOAT, FLOAT, UINT, UINT, POINTER(IDWriteRenderingParams))),
        ("RegisterFontFileLoader",
         com.STDMETHOD(c_void_p)),  # Ambiguous as newer is a pIUnknown and legacy is IUnknown.
        ("UnregisterFontFileLoader",
         com.STDMETHOD(POINTER(IDWriteFontFileLoader_LI))),
        ("CreateTextFormat",
         com.STDMETHOD(c_wchar_p, IDWriteFontCollection, DWRITE_FONT_WEIGHT, DWRITE_FONT_STYLE, DWRITE_FONT_STRETCH,
                       FLOAT, c_wchar_p, POINTER(IDWriteTextFormat))),
        ("CreateTypography",
         com.STDMETHOD(POINTER(IDWriteTypography))),
        ("GetGdiInterop",
         com.STDMETHOD(POINTER(IDWriteGdiInterop))),
        ("CreateTextLayout",
         com.STDMETHOD(c_wchar_p, UINT32, IDWriteTextFormat, FLOAT, FLOAT, POINTER(IDWriteTextLayout))),
        ("CreateGdiCompatibleTextLayout",
         com.STDMETHOD()),
        ("CreateEllipsisTrimmingSign",
         com.STDMETHOD()),
        ("CreateTextAnalyzer",
         com.STDMETHOD(POINTER(IDWriteTextAnalyzer))),
        ("CreateNumberSubstitution",
         com.STDMETHOD()),
        ("CreateGlyphRunAnalysis",
         com.STDMETHOD()),
    ]

# Windows 8
IID_IDWriteFactory1 = com.GUID(0x30572f99, 0xdac6, 0x41db, 0xa1, 0x6e, 0x04, 0x86, 0x30, 0x7e, 0x60, 0x6a)


class IDWriteFactory1(IDWriteFactory, com.pIUnknown):
    _methods_ = [
        ("GetEudcFontCollection",
         com.STDMETHOD()),
        ("CreateCustomRenderingParams1",
         com.STDMETHOD()),
    ]


class DWRITE_UNICODE_RANGE(Structure):
    _fields_ = [
        ("first", UINT32),
        ("last", UINT32)
    ]


class IDWriteFontFallback(com.pIUnknown):
    _methods_ = [
        ("MapCharacters",
         com.STDMETHOD(POINTER(IDWriteTextAnalysisSource), UINT32, UINT32, IDWriteFontCollection, c_wchar_p,
                       DWRITE_FONT_WEIGHT, DWRITE_FONT_STYLE, DWRITE_FONT_STRETCH, POINTER(UINT32),
                       POINTER(IDWriteFont),
                       POINTER(FLOAT))),
    ]

class IDWriteFontFallbackBuilder(com.pIUnknown):
    _methods_ = [
        ("AddMapping",
         com.STDMETHOD(POINTER(DWRITE_UNICODE_RANGE), UINT32, POINTER(c_wchar_p), UINT32,
                       IDWriteFontCollection,
                       c_wchar_p, c_wchar_p, FLOAT)),
        ("AddMappings",
         com.STDMETHOD(IDWriteFontFallback)),
        ("CreateFontFallback",
         com.STDMETHOD(POINTER(IDWriteFontFallback))),
    ]



class IDWriteColorGlyphRunEnumerator(com.pIUnknown):
    _methods_ = [
        ("MoveNext",
         com.STDMETHOD(POINTER(BOOL))),
        ("GetCurrentRun",
         com.STDMETHOD(POINTER(POINTER(DWRITE_COLOR_GLYPH_RUN)))),
    ]


class IDWriteFactory2(IDWriteFactory1, IDWriteFactory, com.pIUnknown):
    _methods_ = [
        ("GetSystemFontFallback",
         com.STDMETHOD(POINTER(IDWriteFontFallback))),
        ("CreateFontFallbackBuilder",
         com.STDMETHOD(POINTER(IDWriteFontFallbackBuilder))),
        ("TranslateColorGlyphRun",
         com.STDMETHOD(FLOAT, FLOAT, POINTER(DWRITE_GLYPH_RUN), c_void_p, DWRITE_MEASURING_MODE, c_void_p, UINT32,
                       POINTER(IDWriteColorGlyphRunEnumerator))),
        ("CreateCustomRenderingParams2",
         com.STDMETHOD()),
        ("CreateGlyphRunAnalysis2",
         com.STDMETHOD()),
    ]

# Windows 8.1
IID_IDWriteFactory2 = com.GUID(0x0439fc60, 0xca44, 0x4994, 0x8d, 0xee, 0x3a, 0x9a, 0xf7, 0xb7, 0x32, 0xec)


class IDWriteFontSet(com.pIUnknown):
    _methods_ = [
        ("GetFontCount",
         com.STDMETHOD()),
        ("GetFontFaceReference",
         com.STDMETHOD()),
        ("FindFontFaceReference",
         com.STDMETHOD()),
        ("FindFontFace",
         com.STDMETHOD()),
        ("GetPropertyValues__",
         com.STDMETHOD()),
        ("GetPropertyValues_",
         com.STDMETHOD()),
        ("GetPropertyValues",
         com.STDMETHOD()),
        ("GetPropertyOccurrenceCount",
         com.STDMETHOD()),
        ("GetMatchingFonts_",
         com.STDMETHOD()),
        ("GetMatchingFonts",
         com.STDMETHOD()),
    ]


class IDWriteFontSetBuilder(com.pIUnknown):
    _methods_ = [
        ("AddFontFaceReference_",
         com.STDMETHOD()),
        ("AddFontFaceReference",
         com.STDMETHOD()),
        ("AddFontSet",
         com.STDMETHOD()),
        ("CreateFontSet",
         com.STDMETHOD(POINTER(IDWriteFontSet))),
    ]


class IDWriteFontSetBuilder1(IDWriteFontSetBuilder, com.pIUnknown):
    _methods_ = [
        ("AddFontFile",
         com.STDMETHOD(IDWriteFontFile)),
    ]


class IDWriteFactory3(IDWriteFactory2, com.pIUnknown):
    _methods_ = [
        ("CreateGlyphRunAnalysis3",
         com.STDMETHOD()),
        ("CreateCustomRenderingParams3",
         com.STDMETHOD()),
        ("CreateFontFaceReference_",
         com.STDMETHOD()),
        ("CreateFontFaceReference",
         com.STDMETHOD()),
        ("GetSystemFontSet",
         com.STDMETHOD()),
        ("CreateFontSetBuilder",
         com.STDMETHOD(POINTER(IDWriteFontSetBuilder))),
        ("CreateFontCollectionFromFontSet",
         com.STDMETHOD(IDWriteFontSet, POINTER(IDWriteFontCollection1))),
        ("GetSystemFontCollection3",
         com.STDMETHOD()),
        ("GetFontDownloadQueue",
         com.STDMETHOD()),
        # ('GetSystemFontSet',
        # com.STDMETHOD()),
    ]


class IDWriteColorGlyphRunEnumerator1(IDWriteColorGlyphRunEnumerator):
    _methods_ = [
        ("GetCurrentRun1",
         com.STDMETHOD(POINTER(POINTER(DWRITE_COLOR_GLYPH_RUN1)))),
    ]

IID_IDWriteColorGlyphRunEnumerator1 = com.GUID(0x7c5f86da, 0xc7a1, 0x4f05, 0xb8,0xe1, 0x55,0xa1,0x79,0xfe,0x5a,0x35)


class IDWriteFactory4(IDWriteFactory3, com.pIUnknown):
    _methods_ = [
        ("TranslateColorGlyphRun4",
         com.STDMETHOD(D2D_POINT_2F, POINTER(DWRITE_GLYPH_RUN), c_void_p, DWRITE_GLYPH_IMAGE_FORMATS,
                       DWRITE_MEASURING_MODE, c_void_p, UINT32, POINTER(IDWriteColorGlyphRunEnumerator1))),
        ("ComputeGlyphOrigins_",
         com.STDMETHOD()),
        ("ComputeGlyphOrigins",
         com.STDMETHOD()),
    ]


class IDWriteInMemoryFontFileLoader(com.pIUnknown):
    _methods_ = [
        ("CreateStreamFromKey",
         com.STDMETHOD()),
        ("CreateInMemoryFontFileReference",
         com.STDMETHOD(IDWriteFactory, c_void_p, UINT, c_void_p, POINTER(IDWriteFontFile))),
        ("GetFileCount",
         com.STDMETHOD()),
    ]

# Windows 10 - Creators
IID_IDWriteFactory5 = com.GUID(0x958DB99A, 0xBE2A, 0x4F09, 0xAF, 0x7D, 0x65, 0x18, 0x98, 0x03, 0xD1, 0xD3)


class IDWriteFactory5(IDWriteFactory4, IDWriteFactory3, IDWriteFactory2, IDWriteFactory1, IDWriteFactory,
                      com.pIUnknown):
    _methods_ = [
        ("CreateFontSetBuilder5",
         com.STDMETHOD(POINTER(IDWriteFontSetBuilder1))),
        ("CreateInMemoryFontFileLoader",
         com.STDMETHOD(POINTER(IDWriteInMemoryFontFileLoader))),
        ("CreateHttpFontFileLoader",
         com.STDMETHOD()),
        ("AnalyzeContainerType",
         com.STDMETHOD()),
        ("UnpackFontFile",
         com.STDMETHOD()),
    ]


ID_IDWriteFactory6 = com.GUID(0xf3744d80, 0x21f7, 0x42eb, 0xb3,0x5d, 0x99,0x5b,0xc7,0x2f,0xc2,0x23)

class IDWriteFactory6(IDWriteFactory5, IDWriteFactory4, IDWriteFactory3, IDWriteFactory2, IDWriteFactory1, IDWriteFactory,
                      com.pIUnknown):
    _methods_ = [
        ("CreateFontFaceReference6",
         com.STDMETHOD()),
        ("CreateFontResource",
         com.STDMETHOD()),
        ("GetSystemFontSet6",
         com.STDMETHOD()),
        ("GetSystemFontCollection6",
         com.STDMETHOD()),
        ("CreateFontCollectionFromFontSet6",
         com.STDMETHOD()),
        ("CreateFontSetBuilder6",
         com.STDMETHOD()),
        ("CreateTextFormat6",
         com.STDMETHOD()),
    ]


class IDWriteFactory7(IDWriteFactory6, IDWriteFactory5, IDWriteFactory4, IDWriteFactory3, IDWriteFactory2,
                      IDWriteFactory1, IDWriteFactory, com.pIUnknown):
    _methods_ = [
        ("GetSystemFontSet7",
         com.STDMETHOD()),
        ("GetSystemFontCollection7",
         com.STDMETHOD()),
    ]

IID_IDWriteFactory7 = com.GUID(0x35d0e0b3, 0x9076, 0x4d2e, 0xa0,0x16, 0xa9,0x1b,0x56,0x8a,0x06,0xb4)


DWriteCreateFactory = dwrite_lib.DWriteCreateFactory
DWriteCreateFactory.restype = HRESULT
DWriteCreateFactory.argtypes = [DWRITE_FACTORY_TYPE, com.REFIID, POINTER(com.pIUnknown)]


# ---- COM Interface implementations.
class Run:
    def __init__(self) -> None:
        self.text_start = 0
        self.text_length = 0
        self.glyph_start = 0
        self.glyph_count = 0
        self.script = DWRITE_SCRIPT_ANALYSIS()
        self.bidi = 0
        self.isNumberSubstituted = False
        self.isSideways = False

        self.next_run = None

    def ContainsTextPosition(self, textPosition: int) -> bool:
        return textPosition >= self.text_start and textPosition < self.text_start + self.text_length


class TextAnalysis(com.COMObject):
    _interfaces_ = [IDWriteTextAnalysisSource, IDWriteTextAnalysisSink]

    def __init__(self) -> None:
        super().__init__()
        self._textstart = 0
        self._textlength = 0
        self._glyphstart = 0
        self._glyphcount = 0
        self._ptrs = []

        self._script = None
        self._bidi = 0
        # self._sideways = False  # noqa: ERA001

    def GenerateResults(self, analyzer: IDWriteTextAnalyzer, text: c_wchar_p, text_length: int) -> None:
        self._text = text
        self._textstart = 0
        self._textlength = text_length
        self._glyphstart = 0
        self._glyphcount = 0
        self._ptrs.clear()

        self._start_run = Run()
        self._start_run.text_length = text_length

        self._current_run = self._start_run

        analyzer.AnalyzeScript(self, 0, text_length, self)

    def SetScriptAnalysis(self, textPosition: UINT32, textLength: UINT32,
                          scriptAnalysis: POINTER(DWRITE_SCRIPT_ANALYSIS)) -> int:
        # textPosition - The index of the first character in the string that the result applies to
        # textLength - How many characters of the string from the index that the result applies to
        # scriptAnalysis - The analysis information for all glyphs starting at position for length.
        self.SetCurrentRun(textPosition)
        self.SplitCurrentRun(textPosition)

        while textLength > 0:
            run, textLength = self.FetchNextRun(textLength)

            run.script.script = scriptAnalysis[0].script
            run.script.shapes = scriptAnalysis[0].shapes

            self._script = run.script

        return 0
        # return 0x80004001

    def GetTextBeforePosition(self, textPosition: UINT32, textString: POINTER(POINTER(WCHAR)),
                              textLength: POINTER(UINT32)) -> NoReturn:
        msg = "Currently not implemented."
        raise Exception(msg)

    def GetTextAtPosition(self, textPosition: UINT32, textString: c_wchar_p, textLength: POINTER(UINT32)) -> int:
        # This method will retrieve a substring of the text in this layout
        #   to be used in an analysis step.
        # Arguments:
        # textPosition - The index of the first character of the text to retrieve.
        # textString - The pointer to the first character of text at the index requested.
        # textLength - The characters available at/after the textString pointer (string length).

        if textPosition >= self._textlength:
            self._no_ptr = c_wchar_p(None)
            textString[0] = self._no_ptr
            textLength[0] = 0
        else:
            ptr = c_wchar_p(self._text[textPosition:])
            self._ptrs.append(ptr)
            textString[0] = ptr
            textLength[0] = self._textlength - textPosition

        return 0

    def GetParagraphReadingDirection(self) -> int:
        return 0

    def GetLocaleName(self, textPosition: UINT32, textLength: POINTER(UINT32),
                      localeName: POINTER(POINTER(WCHAR))) -> int:
        self.__local_name = c_wchar_p("")  # TODO: Add more locales.
        localeName[0] = self.__local_name
        textLength[0] = self._textlength - textPosition
        return 0

    def GetNumberSubstitution(self) -> int:
        return 0

    def SetCurrentRun(self, textPosition: UINT32) -> None:
        if self._current_run and self._current_run.ContainsTextPosition(textPosition):
            return

    def SplitCurrentRun(self, textPosition: UINT32) -> None:
        if not self._current_run:
            return

        if textPosition <= self._current_run.text_start:
            # Already first start of the run.
            return

        new_run = copy.copy(self._current_run)

        new_run.next_run = self._current_run.next_run
        self._current_run.next_run = new_run

        splitPoint = textPosition - self._current_run.text_start
        new_run.text_start += splitPoint
        new_run.text_length -= splitPoint

        self._current_run.text_length = splitPoint
        self._current_run = new_run

    def FetchNextRun(self, textLength: UINT32) -> tuple[Run, int]:
        original_run = self._current_run

        if (textLength < self._current_run.text_length):
            self.SplitCurrentRun(self._current_run.text_start + textLength)
        else:
            self._current_run = self._current_run.next_run

        textLength -= original_run.text_length

        return original_run, textLength


class MyFontFileStream(com.COMObject):
    _interfaces_ = [IDWriteFontFileStream]

    def __init__(self, data: bytes) -> None:
        super().__init__()
        self._data = data
        self._size = len(data)
        self._ptrs = []

    def ReadFileFragment(self, fragmentStart: POINTER(c_void_p), fileOffset: UINT64, fragmentSize: UINT64,
                         fragmentContext: POINTER(c_void_p)) -> int:
        if fileOffset + fragmentSize > self._size:
            return 0x80004005  # E_FAIL

        fragment = self._data[fileOffset:]
        buffer = (ctypes.c_ubyte * len(fragment)).from_buffer(bytearray(fragment))
        ptr = ctypes.cast(buffer, c_void_p)

        self._ptrs.append(ptr)
        fragmentStart[0] = ptr
        fragmentContext[0] = None
        return 0

    def ReleaseFileFragment(self, fragmentContext: c_void_p) -> int:
        return 0

    def GetFileSize(self, fileSize: POINTER(UINT64)) -> int:
        fileSize[0] = self._size
        return 0

    def GetLastWriteTime(self, lastWriteTime: POINTER(UINT64)) -> int:
        return com.E_NOTIMPL


class LegacyFontFileLoader(com.COMObject):
    _interfaces_ = [IDWriteFontFileLoader_LI]

    def __init__(self) -> None:
        super().__init__()
        self._streams = {}

    def CreateStreamFromKey(self, fontfileReferenceKey: c_void_p, fontFileReferenceKeySize: UINT32,
                            fontFileStream: POINTER(IDWriteFontFileStream)) -> int:
        convert_index = cast(fontfileReferenceKey, POINTER(c_uint32))

        self._ptr = cast(self._streams[convert_index.contents.value].as_interface(IDWriteFontFileStream),
                         POINTER(IDWriteFontFileStream))
        fontFileStream[0] = self._ptr
        return 0

    def SetCurrentFont(self, index: int, data: bytes) -> int:
        self._streams[index] = MyFontFileStream(data)


class MyEnumerator(com.COMObject):
    _interfaces_ = [IDWriteFontFileEnumerator]

    def __init__(self, factory: c_void_p, loader: LegacyFontFileLoader) -> None:
        super().__init__()
        self.factory = ctypes.cast(factory, IDWriteFactory)
        self.key = "pyglet_dwrite"
        self.size = len(self.key)
        self.current_index = -1

        self._keys = []
        self._font_data = []
        self._font_files = []
        self._current_file = None

        self._font_key_ref = create_unicode_buffer("none")
        self._font_key_len = len(self._font_key_ref)

        self._file_loader = loader

    def AddFontData(self, fonts: list[str]) -> None:
        self._font_data = fonts

    def MoveNext(self, hasCurrentFile: BOOL) -> None:
        self.current_index += 1
        if self.current_index != len(self._font_data):
            font_file = IDWriteFontFile()

            self._file_loader.SetCurrentFont(self.current_index, self._font_data[self.current_index])

            key = self.current_index

            if self.current_index not in self._keys:
                buffer = pointer(c_uint32(key))

                ptr = cast(buffer, c_void_p)

                self._keys.append(ptr)

                self.factory.CreateCustomFontFileReference(self._keys[self.current_index],
                                                           sizeof(buffer),
                                                           self._file_loader,
                                                           byref(font_file))

                self._font_files.append(font_file)

                hasCurrentFile[0] = 1
            else:
                hasCurrentFile[0] = 0
        else:
            hasCurrentFile[0] = 0

    def GetCurrentFontFile(self, fontFile: IDWriteFontFile) -> int:
        fontFile = cast(fontFile, POINTER(IDWriteFontFile))
        fontFile[0] = self._font_files[self.current_index]
        return 0


class LegacyCollectionLoader(com.COMObject):
    _interfaces_ = [IDWriteFontCollectionLoader]

    def __init__(self, factory: c_void_p, loader: LegacyFontFileLoader) -> None:
        super().__init__()
        self._enumerator = MyEnumerator(factory, loader)

    def AddFontData(self, fonts) -> None:
        self._enumerator.AddFontData(fonts)

    def CreateEnumeratorFromKey(self, factory: IDWriteFactory, key: c_void_p, key_size: UINT32,
                                enumerator: MyEnumerator) -> int:
        self._ptr = cast(self._enumerator.as_interface(IDWriteFontFileEnumerator),
                         POINTER(IDWriteFontFileEnumerator))

        enumerator[0] = self._ptr
        return 0



from __future__ import annotations

from unittest.mock import MagicMock, patch

import pytest

import pyglet


@pytest.fixture(autouse=True)
def monkeypatch_default_sprite_shader(monkeypatch, get_dummy_shader_program):
    """Use a dummy shader when testing non-drawing functionality"""
    monkeypatch.setattr('pyglet.sprite.get_default_shader', get_dummy_shader_program)


@pytest.fixture()
def sprite():
    """A sprite with no image data.

    It is created at a non-zero position so that the update method can
    be verified as working when passed zero positions (Issue #673).
    Scale values do not need to be set to 0 in the fixture as they are
    all 1.0 by default.

    The lack of image data doesn't matter because these tests never touch
    a real GL context which would require it.
    """
    sprite = pyglet.sprite.Sprite(MagicMock(), x=1, y=2, z=3)
    sprite.rotation = 90

    return sprite


@pytest.fixture()
def batched_sprite():
    """A sprite with no image data, but is initially added to a batch.

    It is created at a non-zero position so that the update method can
    be verified as working when passed zero positions (Issue #673).
    Scale values do not need to be set to 0 in the fixture as they are
    all 1.0 by default.

    The lack of image data doesn't matter because these tests never touch
    a real GL context which would require it.
    """
    from pyglet.graphics import Batch
    sprite = pyglet.sprite.Sprite(MagicMock(), x=1, y=2, z=3, batch=Batch())
    sprite.rotation = 90

    return sprite


def _new_or_none(new_value: float = 0.0) -> tuple[float, None]:
    return new_value, None


@pytest.fixture(params=_new_or_none())
def x(request):
    return request.param


@pytest.fixture(params=_new_or_none())
def y(request):
    return request.param


@pytest.fixture(params=_new_or_none())
def z(request):
    return request.param


@pytest.fixture(params=_new_or_none())
def scale(request):
    return request.param


@pytest.fixture(params=_new_or_none())
def scale_x(request):
    return request.param


@pytest.fixture(params=_new_or_none())
def scale_y(request):
    return request.param


def test_color_initially_rgba_white(sprite):
    assert sprite.color == (255, 255, 255, 255)


def test_opacity_initially_255(sprite):
    assert sprite.opacity == 255

def test_color_changes_opacity_if_set_to_rgba(sprite, new_rgba_color):
    sprite.color = new_rgba_color
    assert sprite.opacity == new_rgba_color[3]


def test_color_leaves_opacity_unchanged_if_set_to_rgb(sprite, new_rgb_color):
    sprite.color = new_rgb_color
    assert sprite.opacity == 255


def test_color_changes_rgb_channels(sprite, new_rgb_or_rgba_color):
    sprite.color = new_rgb_or_rgba_color
    assert sprite.color[:3] == new_rgb_or_rgba_color[:3]


def test_opacity_setter_changes_opacity_channel_in_color(sprite, new_opacity_set_opacity_alone):
    sprite.opacity = new_opacity_set_opacity_alone
    assert sprite.color[3] == new_opacity_set_opacity_alone


def test_update_sets_passed_positions(sprite, x, y, z):
    sprite.update(x=x, y=y, z=z)

    if x is not None:
        assert sprite.x == x

    if y is not None:
        assert sprite.y == y

    if z is not None:
        assert sprite.z == z


def test_update_leaves_none_translations_alone(sprite, x, y, z):
    o_x, o_y, o_z = sprite.x, sprite.y, sprite.z

    sprite.update(x=x, y=y, z=z)

    if x is None:
        assert sprite.x == o_x

    if y is None:
        assert sprite.y == o_y

    if z is None:
        assert sprite.z == o_z


def test_update_sets_passed_scales(sprite, scale, scale_x, scale_y):
    sprite.update(scale=scale, scale_x=scale_x, scale_y=scale_y)

    if scale is not None:
        assert sprite.scale == scale

    if scale_x is not None:
        assert sprite.scale_x == scale_x

    if scale_y is not None:
        assert sprite.scale_y == scale_y


def test_update_leaves_none_scales_alone(sprite, scale, scale_x, scale_y):
    o_scale, o_scale_x, o_scale_y = sprite.scale, sprite.scale_x, sprite.scale_y

    sprite.update(scale=scale, scale_x=scale_x, scale_y=scale_y)

    if scale is None:
        assert sprite.scale == o_scale

    if scale_x is None:
        assert sprite.scale_x == o_scale_x

    if scale_y is None:
        assert sprite.scale_y == o_scale_y


def test_update_sets_rotation_when_passed(sprite):
    sprite.update(rotation=0.0)
    assert sprite.rotation == 0.0


def test_update_leaves_rotation_alone_when_none(sprite):
    sprite.update()
    assert sprite.rotation == 90


@pytest.mark.parametrize('fixture', ['sprite', 'batched_sprite'])
def test_group_setter(request, fixture):
    _sprite = request.getfixturevalue(fixture)
    from pyglet.graphics import Group
    new_group = Group()
    # Patch since magicmock returns functions for class variables...
    with patch.multiple(_sprite._vertex_list, indexed=True, instanced=False):  # noqa: SLF001
        _sprite.group = new_group

        assert _sprite.group is new_group


@pytest.mark.parametrize('fixture', ['sprite', 'batched_sprite'])
def test_batch_setter(request, fixture):
    _sprite = request.getfixturevalue(fixture)
    from pyglet.graphics import Batch
    new_batch = Batch()
    with patch.multiple(_sprite._vertex_list, indexed=True, instanced=False):  # noqa: SLF001
        _sprite.batch = new_batch
        assert _sprite.batch is new_batch


@pytest.mark.parametrize('fixture', ['sprite', 'batched_sprite'])
def test_program_setter(request, fixture):
    _sprite = request.getfixturevalue(fixture)

    program = MagicMock()
    with patch.multiple(_sprite._vertex_list, indexed=True, instanced=False):  # noqa: SLF001
        _sprite.program = program
        assert _sprite.program == program


@pytest.mark.parametrize('fixture', ['sprite', 'batched_sprite'])
def test_blend_setter(request, fixture):
    _sprite = request.getfixturevalue(fixture)

    blend_mode = (1, 1)
    with patch.multiple(_sprite._vertex_list, indexed=True, instanced=False):  # noqa: SLF001
        _sprite.blend_mode = blend_mode
        assert _sprite._group.blend_src == 1  # noqa: SLF001
        assert _sprite._group.blend_dest == 1  # noqa: SLF001



from __future__ import annotations

import sys

import pyglet
from pyglet import clock, event, graphics, image
from pyglet.gl import *

_is_pyglet_doc_run = hasattr(sys, "is_pyglet_doc_run") and sys.is_pyglet_doc_run

vertex_source = """#version 150
    in vec3 position;
    in vec4 size;
    in vec2 scale;
    in vec4 color;
    in vec4 texture_uv;
    in float rotation;

    out vec4 geo_size;
    out vec2 geo_scale;
    out vec4 geo_color;
    out vec4 geo_tex_coords;
    out float geo_rotation;

    void main() {
        gl_Position = vec4(position, 1);
        geo_size = size;
        geo_scale = scale;
        geo_color = color;
        geo_tex_coords = texture_uv;
        geo_rotation = rotation;
    }
"""

geometry_source = """#version 150
    // We are taking single points from the vertex shader
    // and emitting 4 new vertices to create a quad/sprite.
    layout (points) in;
    layout (triangle_strip, max_vertices = 4) out;

    uniform WindowBlock
    {
        mat4 projection;
        mat4 view;
    } window;


    // Since geometry shader can take multiple values from a vertex
    // shader we need to define the inputs from it as arrays.
    // For our purposes, we just take single values (points).
    in vec4 geo_size[];
    in vec2 geo_scale[];
    in vec4 geo_color[];
    in vec4 geo_tex_coords[];
    in float geo_rotation[];

    out vec2 uv;
    out vec4 frag_color;

    void main() {

        // Unpack the image size and anchor
        vec2 size = geo_size[0].xy;
        vec2 anchor = geo_size[0].zw;

        // Grab the position value from the vertex shader
        vec3 center = gl_in[0].gl_Position.xyz;

        // This matrix controls the actual position of the sprite
        mat4 m_translate = mat4(1.0);
        m_translate[3][0] = center.x;
        m_translate[3][1] = center.y;
        m_translate[3][2] = center.z;

        mat4 m_rotation = mat4(1.0);
        m_rotation[0][0] =  cos(radians(-geo_rotation[0])); 
        m_rotation[0][1] =  sin(radians(-geo_rotation[0]));
        m_rotation[1][0] = -sin(radians(-geo_rotation[0]));
        m_rotation[1][1] =  cos(radians(-geo_rotation[0]));

        mat4 m_scale = mat4(1.0);
        m_scale[0][0] = geo_scale[0].x;
        m_scale[1][1] = geo_scale[0].y;

        // Final UV coords (left, bottom, right, top):
        float uv_l = geo_tex_coords[0].s;
        float uv_b = geo_tex_coords[0].t;
        float uv_r = geo_tex_coords[0].p;
        float uv_t = geo_tex_coords[0].q;

        // Emit a triangle strip to create a quad (4 vertices).
        // Prepare and reuse the transformation matrix and fragment color:
        mat4 m_pv = window.projection * window.view * m_translate * m_rotation * m_scale;
        frag_color = geo_color[0];

        // Upper left
        gl_Position = m_pv * vec4(vec2(0.0, size.y) - anchor, 0.0, 1.0);
        uv = vec2(uv_l, uv_t);
        EmitVertex();

        // lower left
        gl_Position = m_pv * vec4(vec2(0.0, 0.0) - anchor, 0.0, 1.0);
        uv = vec2(uv_l, uv_b);
        EmitVertex();

        // upper right
        gl_Position = m_pv * vec4(vec2(size.x, size.y) - anchor, 0.0, 1.0);
        uv = vec2(uv_r, uv_t);
        EmitVertex();

        // lower right
        gl_Position = m_pv * vec4(vec2(size.x, 0.0) - anchor, 0.0, 1.0);
        uv = vec2(uv_r, uv_b);
        EmitVertex();

        // We are done with this triangle strip now
        EndPrimitive();
    }
"""

fragment_source = """#version 150
    in vec2 uv;
    in vec4 frag_color;
    out vec4 final_color;

    uniform sampler2D sprite_texture;

    void main() {
        final_color = texture(sprite_texture, uv) * frag_color;
    }

"""

fragment_array_source = """#version 150 core
    in vec2 uv;
    in vec4 frag_color;
    
    out vec4 final_colors;

    uniform sampler2DArray sprite_texture;

    void main()
    {
        final_colors = texture(sprite_texture, uv) * frag_color;
    }
"""


def get_default_shader():
    return pyglet.gl.current_context.create_program((vertex_source, 'vertex'),
                                                    (geometry_source, 'geometry'),
                                                    (fragment_source, 'fragment'))


def get_default_array_shader():
    return pyglet.gl.current_context.create_program((vertex_source, 'vertex'),
                                                    (geometry_source, 'geometry'),
                                                    (fragment_array_source, 'fragment'))


class SpriteGroup(graphics.Group):
    """Shared sprite rendering group.

    The group is automatically coalesced with other sprite groups sharing the
    same parent group, texture and blend parameters.
    """

    def __init__(self, texture, blend_src, blend_dest, program, parent=None):
        """Create a sprite group.

        The group is created internally when a :py:class:`~pyglet.sprite.Sprite`
        is created; applications usually do not need to explicitly create it.

        :Parameters:
            `texture` : `~pyglet.image.Texture`
                The (top-level) texture containing the sprite image.
            `blend_src` : int
                OpenGL blend source mode; for example,
                ``GL_SRC_ALPHA``.
            `blend_dest` : int
                OpenGL blend destination mode; for example,
                ``GL_ONE_MINUS_SRC_ALPHA``.
            `program` : `~pyglet.graphics.shader.ShaderProgram`
                A custom ShaderProgram.
            `order` : int
                Change the order to render above or below other Groups.
            `parent` : `~pyglet.graphics.Group`
                Optional parent group.
        """
        super().__init__(parent=parent)
        self.texture = texture
        self.blend_src = blend_src
        self.blend_dest = blend_dest
        self.program = program

    def set_state(self):
        self.program.use()

        glActiveTexture(GL_TEXTURE0)
        glBindTexture(self.texture.target, self.texture.id)

        glEnable(GL_BLEND)
        glBlendFunc(self.blend_src, self.blend_dest)

    def unset_state(self):
        glDisable(GL_BLEND)
        self.program.stop()

    def __repr__(self):
        return f"{self.__class__.__name__}({self.texture})"

    def __eq__(self, other):
        return (other.__class__ is self.__class__ and
                self.program is other.program and
                self.parent == other.parent and
                self.texture.target == other.texture.target and
                self.texture.id == other.texture.id and
                self.blend_src == other.blend_src and
                self.blend_dest == other.blend_dest)

    def __hash__(self):
        return hash((self.program, self.parent,
                     self.texture.id, self.texture.target,
                     self.blend_src, self.blend_dest))


class Sprite(event.EventDispatcher):
    _batch = None
    _animation = None
    _frame_index = 0
    _paused = False
    _rotation = 0
    _scale = 1.0
    _scale_x = 1.0
    _scale_y = 1.0
    _visible = True
    _vertex_list = None
    group_class = SpriteGroup

    def __init__(self,
                 img, x=0, y=0, z=0,
                 blend_src=GL_SRC_ALPHA, blend_dest=GL_ONE_MINUS_SRC_ALPHA,
                 batch=None, group=None, subpixel=False, program=None):
        """Create a sprite.

        :Parameters:
            `img` : `~pyglet.image.AbstractImage` or `~pyglet.image.Animation`
                Image or animation to display.
            `x` : int
                X coordinate of the sprite.
            `y` : int
                Y coordinate of the sprite.
            `z` : int
                Z coordinate of the sprite.
            `blend_src` : int
                OpenGL blend source mode.  The default is suitable for
                compositing sprites drawn from back-to-front.
            `blend_dest` : int
                OpenGL blend destination mode.  The default is suitable for
                compositing sprites drawn from back-to-front.
            `batch` : `~pyglet.graphics.Batch`
                Optional batch to add the sprite to.
            `group` : `~pyglet.graphics.Group`
                Optional parent group of the sprite.
            `subpixel` : bool
                Allow floating-point coordinates for the sprite. By default,
                coordinates are restricted to integer values.
            `program` : `~pyglet.graphics.shader.ShaderProgram`
                A custom shader to use. This shader program must contain the
                exact same attribute names and types as the default shader.
                The class methods and properties depend on this, and will
                crash otherwise.
        """
        self._x = x
        self._y = y
        self._z = z
        self._img = img

        self._rgba = [255, 255, 255, 255]

        if isinstance(img, image.Animation):
            self._animation = img
            self._texture = img.frames[0].image.get_texture()
            self._next_dt = img.frames[0].duration
            if self._next_dt:
                clock.schedule_once(self._animate, self._next_dt)
        else:
            self._texture = img.get_texture()

        if not program:
            if isinstance(img, image.TextureArrayRegion):
                self._program = get_default_array_shader()
            else:
                self._program = get_default_shader()
        else:
            self._program = program

        self._batch = batch or graphics.get_default_batch()
        self._user_group = group
        self._group = self.group_class(self._texture, blend_src, blend_dest, self.program, group)
        self._subpixel = subpixel

        self._create_vertex_list()

    def _create_vertex_list(self):
        texture = self._texture
        self._vertex_list = self.program.vertex_list(
            1, GL_POINTS, self._batch, self._group,
            position=('f', (self._x, self._y, self._z)),
            size=('f', (texture.width, texture.height, texture.anchor_x, texture.anchor_y)),
            scale=('f', (self._scale_x, self._scale_y)),
            color=('Bn', self._rgba),
            texture_uv=('f', texture.uv),
            rotation=('f', (self._rotation,)))

    @property
    def program(self):
        return self._program

    @program.setter
    def program(self, program):
        if self._program == program:
            return
        self._group = self.group_class(self._texture,
                                       self._group.blend_src,
                                       self._group.blend_dest,
                                       program,
                                       self._user_group)
        if (self._batch and
                self._batch.update_shader(self._vertex_list, GL_POINTS, self._group, program)):
            # Exit early if changing domain is not needed.
            return

        # Recreate vertex list.
        self._vertex_list.delete()
        self._create_vertex_list()

    def delete(self):
        """Force immediate removal of the sprite from video memory.

        This is often necessary when using batches, as the Python garbage
        collector will not necessarily call the finalizer as soon as the
        sprite is garbage.
        """
        if self._animation:
            clock.unschedule(self._animate)
        self._vertex_list.delete()
        self._vertex_list = None
        self._texture = None
        self._group = None

    def _animate(self, dt):
        self._frame_index += 1
        if self._frame_index >= len(self._animation.frames):
            self._frame_index = 0
            self.dispatch_event('on_animation_end')
            if self._vertex_list is None:
                return  # Deleted in event handler.

        frame = self._animation.frames[self._frame_index]
        self._set_texture(frame.image.get_texture())

        if frame.duration is not None:
            duration = frame.duration - (self._next_dt - dt)
            duration = min(max(0, duration), frame.duration)
            clock.schedule_once(self._animate, duration)
            self._next_dt = duration
        else:
            self.dispatch_event('on_animation_end')

    @property
    def batch(self):
        """Graphics batch.

        The sprite can be migrated from one batch to another, or removed from
        its batch (for individual drawing).  Note that this can be an expensive
        operation.

        :type: :py:class:`pyglet.graphics.Batch`
        """
        return self._batch

    @batch.setter
    def batch(self, batch):
        if self._batch == batch:
            return

        if batch is not None and self._batch is not None:
            self._batch.migrate(self._vertex_list, GL_POINTS, self._group, batch)
            self._batch = batch
        else:
            self._vertex_list.delete()
            self._batch = batch
            self._create_vertex_list()

    @property
    def group(self):
        """Parent graphics group.

        The sprite can change its rendering group, however this can be an
        expensive operation.

        :type: :py:class:`pyglet.graphics.Group`
        """
        return self._group.parent

    @group.setter
    def group(self, group):
        if self._group.parent == group:
            return
        self._group = self.group_class(self._texture,
                                       self._group.blend_src,
                                       self._group.blend_dest,
                                       self._group.program,
                                       group)
        self._batch.migrate(self._vertex_list, GL_POINTS, self._group, self._batch)

    @property
    def image(self):
        """Image or animation to display.

        :type: :py:class:`~pyglet.image.AbstractImage` or
               :py:class:`~pyglet.image.Animation`
        """
        if self._animation:
            return self._animation
        return self._texture

    @image.setter
    def image(self, img):
        if self._animation is not None:
            clock.unschedule(self._animate)
            self._animation = None

        if isinstance(img, image.Animation):
            self._animation = img
            self._frame_index = 0
            self._set_texture(img.frames[0].image.get_texture())
            self._next_dt = img.frames[0].duration
            if self._next_dt:
                clock.schedule_once(self._animate, self._next_dt)
        else:
            self._set_texture(img.get_texture())

    def _set_texture(self, texture):
        if texture.id is not self._texture.id:
            self._group = self._group.__class__(texture,
                                                self._group.blend_src,
                                                self._group.blend_dest,
                                                self._group.program,
                                                self._group.parent)
            self._vertex_list.delete()
            self._texture = texture
            self._create_vertex_list()
        else:
            self._vertex_list.texture_uv[:] = texture.uv
        self._texture = texture

    @property
    def position(self):
        """The (x, y, z) coordinates of the sprite, as a tuple.

        :Parameters:
            `x` : int
                X coordinate of the sprite.
            `y` : int
                Y coordinate of the sprite.
            `z` : int
                Z coordinate of the sprite.
        """
        return self._x, self._y, self._z

    @position.setter
    def position(self, position):
        self._x, self._y, self._z = position
        self._vertex_list.position[:] = position

    @property
    def x(self):
        """X coordinate of the sprite.

        :type: int
        """
        return self._x

    @x.setter
    def x(self, x):
        self._x = x
        self._vertex_list.position[:] = x, self._y, self._z

    @property
    def y(self):
        """Y coordinate of the sprite.

        :type: int
        """
        return self._y

    @y.setter
    def y(self, y):
        self._y = y
        self._vertex_list.position[:] = self._x, y, self._z

    @property
    def z(self):
        """Z coordinate of the sprite.

        :type: int
        """
        return self._z

    @z.setter
    def z(self, z):
        self._z = z
        self._vertex_list.position[:] = self._x, self._y, z

    @property
    def rotation(self):
        """Clockwise rotation of the sprite, in degrees.

        The sprite image will be rotated about its image's (anchor_x, anchor_y)
        position.

        :type: float
        """
        return self._rotation

    @rotation.setter
    def rotation(self, rotation):
        self._rotation = rotation
        self._vertex_list.rotation[0] = self._rotation

    @property
    def scale(self):
        """Base Scaling factor.

        A scaling factor of 1 (the default) has no effect.  A scale of 2 will
        draw the sprite at twice the native size of its image.

        :type: float
        """
        return self._scale

    @scale.setter
    def scale(self, scale):
        self._scale = scale
        self._vertex_list.scale[:] = scale * self._scale_x, scale * self._scale_y

    @property
    def scale_x(self):
        """Horizontal scaling factor.

         A scaling factor of 1 (the default) has no effect.  A scale of 2 will
         draw the sprite at twice the native width of its image.

        :type: float
        """
        return self._scale_x

    @scale_x.setter
    def scale_x(self, scale_x):
        self._scale_x = scale_x
        self._vertex_list.scale[:] = self._scale * scale_x, self._scale * self._scale_y

    @property
    def scale_y(self):
        """Vertical scaling factor.

         A scaling factor of 1 (the default) has no effect.  A scale of 2 will
         draw the sprite at twice the native height of its image.

        :type: float
        """
        return self._scale_y

    @scale_y.setter
    def scale_y(self, scale_y):
        self._scale_y = scale_y
        self._vertex_list.scale[:] = self._scale * self._scale_x, self._scale * scale_y

    def update(self, x=None, y=None, z=None, rotation=None, scale=None, scale_x=None, scale_y=None):
        """Simultaneously change the position, rotation or scale.

        This method is provided for convenience. There is not much
        performance benefit to updating multiple Sprite attributes at once.

        :Parameters:
            `x` : int
                X coordinate of the sprite.
            `y` : int
                Y coordinate of the sprite.
            `z` : int
                Z coordinate of the sprite.
            `rotation` : float
                Clockwise rotation of the sprite, in degrees.
            `scale` : float
                Scaling factor.
            `scale_x` : float
                Horizontal scaling factor.
            `scale_y` : float
                Vertical scaling factor.
        """
        translations_outdated = False

        # only bother updating if the translation actually changed
        if x is not None:
            self._x = x
            translations_outdated = True
        if y is not None:
            self._y = y
            translations_outdated = True
        if z is not None:
            self._z = z
            translations_outdated = True

        if translations_outdated:
            self._vertex_list.position[:] = (self._x, self._y, self._z)

        if rotation is not None and rotation != self._rotation:
            self._rotation = rotation
            self._vertex_list.rotation[:] = rotation

        scales_outdated = False

        # only bother updating if the scale actually changed
        if scale is not None:
            self._scale = scale
            scales_outdated = True
        if scale_x is not None:
            self._scale_x = scale_x
            scales_outdated = True
        if scale_y is not None:
            self._scale_y = scale_y
            scales_outdated = True

        if scales_outdated:
            self._vertex_list.scale[:] = self._scale * self._scale_x, self._scale * self._scale_y

    @property
    def width(self):
        """Scaled width of the sprite.

        Invariant under rotation.

        :type: int
        """
        w = self._texture.width * abs(self._scale_x) * abs(self._scale)
        return w if self._subpixel else int(w)

    @width.setter
    def width(self, width):
        self.scale_x = width / (self._texture.width * abs(self._scale))

    @property
    def height(self):
        """Scaled height of the sprite.

        Invariant under rotation.

        :type: int
        """
        h = self._texture.height * abs(self._scale_y) * abs(self._scale)
        return h if self._subpixel else int(h)

    @height.setter
    def height(self, height):
        self.scale_y = height / (self._texture.height * abs(self._scale))

    @property
    def opacity(self):
        """Blend opacity.

        This property sets the alpha component of the colour of the sprite's
        vertices.  With the default blend mode (see the constructor), this
        allows the sprite to be drawn with fractional opacity, blending with the
        background.

        An opacity of 255 (the default) has no effect.  An opacity of 128 will
        make the sprite appear translucent.

        :type: int
        """
        return self._rgba[3]

    @opacity.setter
    def opacity(self, opacity):
        self._rgba[3] = opacity
        self._vertex_list.color[:] = self._rgba

    @property
    def color(self):
        """Blend color.

        This property sets the color of the sprite's vertices. This allows the
        sprite to be drawn with a color tint.

        The color is specified as an RGB tuple of integers '(red, green, blue)'.
        Each color component must be in the range 0 (dark) to 255 (saturated).

        :type: (int, int, int)
        """
        return self._rgba[:3]

    @color.setter
    def color(self, rgb):
        self._rgba[:3] = list(map(int, rgb))
        self._vertex_list.color[:] = self._rgba

    @property
    def visible(self):
        """True if the sprite will be drawn.

        :type: bool
        """
        return self._visible

    @visible.setter
    def visible(self, visible):
        self._visible = visible
        self._vertex_list.texture_uv[:] = (0, 0, 0, 0) if not visible else self._texture.uv

    @property
    def paused(self):
        """Pause/resume the Sprite's Animation

        If `Sprite.image` is an Animation, you can pause or resume
        the animation by setting this property to True or False.
        If not an Animation, this has no effect.

        :type: bool
        """
        return self._paused

    @paused.setter
    def paused(self, pause):
        if not hasattr(self, '_animation') or pause == self._paused:
            return
        if pause is True:
            clock.unschedule(self._animate)
        else:
            frame = self._animation.frames[self._frame_index]
            self._next_dt = frame.duration
            if self._next_dt:
                clock.schedule_once(self._animate, self._next_dt)
        self._paused = pause

    @property
    def frame_index(self):
        """The current Animation frame.

        If the `Sprite.image` is an `Animation`,
        you can query or set the current frame.
        If not an Animation, this will always
        be 0.

        :type: int
        """
        return self._frame_index

    @frame_index.setter
    def frame_index(self, index):
        # Bound to available number of frames
        if self._animation is None:
            return
        self._frame_index = max(0, min(index, len(self._animation.frames) - 1))

    def draw(self):
        """Draw the sprite at its current position.

        See the module documentation for hints on drawing multiple sprites
        efficiently.
        """
        self._group.set_state_recursive()
        self._vertex_list.draw(GL_POINTS)
        self._group.unset_state_recursive()

    def __del__(self):
        try:
            if self._vertex_list is not None:
                self._vertex_list.delete()
        except:
            pass

    if _is_pyglet_doc_run:
        def on_animation_end(self):
            """The sprite animation reached the final frame.

            The event is triggered only if the sprite has an animation, not an
            image.  For looping animations, the event is triggered each time
            the animation loops.

            :event:
            """


Sprite.register_event_type('on_animation_end')



"""Experimental networking

This module contains experiments in making user-friendly Server and Client
classes that integrate with pyglet's event system. These are very basic,
socket server/client examples, and are not ready to be used in production.
They are included here to solicit feedback, and possibly spark further
development. Basic Server usage::

    server = net.Server(address='0.0.0.0', port=1234)
    active_connections = weakref.WeakSet()

    def pong(connection, message):
        print(f"Received '{message}' from '{connection}'")
        connection.send(b'pong')

    @server.event
    def on_connection(connection):
        print(f"New client connected: {connection}")
        connection.set_handler('on_receive', pong)
        active_connections.add(connection)

    @server.event
    def on_disconnect(connection):
        print(f"Client disconnected: {connection}")
        active_connections.discard(connection)


Basic Client example::

    client = net.Client(address='localhost', port=1234)

    @client.event
    def on_receive(client, message):
        print(f"Received: {message}")

    @client.event
    def on_disconnect(client):
        print(f"Disconnected: {client}")

    client.send(b'ping')

"""


import queue as _queue
import struct as _struct
import socket as _socket
import asyncio as _asyncio
import threading as _threading

from pyglet.event import EventDispatcher as _EventDispatcher

from pyglet.util import debug_print

_debug_net = debug_print('debug_net')


class Client(_EventDispatcher):
    def __init__(self, address, port):
        """Create a Client connection to a Server."""
        self._socket = _socket.create_connection((address, port))
        self._address = address
        self._port = port

        self._terminate = _threading.Event()
        self._queue = _queue.Queue()

        _threading.Thread(target=self._recv, daemon=True).start()
        _threading.Thread(target=self._send, daemon=True).start()

        self._sentinal = object()  # poison pill

    def close(self):
        """Close the connection."""
        self._queue.put(self._sentinal)
        self._socket.shutdown(1)
        if not self._terminate.is_set():
            self._terminate.set()
            self.dispatch_event('on_disconnect', self)

    def send(self, message):
        """Queue a message to send.

        Put a string of bytes into the queue to send.
        raises a `ConnectionError` if the connection
        has been closed or dropped.

       :Parameters:
            `message` : bytes
                A string of bytes to send.
        """
        if self._terminate.is_set():
            raise ConnectionError("Connection is closed.")
        self._queue.put(message)

    def _send(self):    # Thread
        """Background Thread to send messages from the queue."""
        while not self._terminate.is_set():
            message = self._queue.get()
            if message == self._sentinal:  # bail out on poison pill
                break
            try:
                # Attach a 4byte header to the front of the message:
                packet = _struct.pack('I', len(message)) + message
                self._socket.sendall(packet)
            except (ConnectionError, OSError):
                self.close()
                break

        assert _debug_net("Exiting _send thread")


    def _recv(self):    # Thread
        socket = self._socket

        while not self._terminate.is_set():
            try:
                header = socket.recv(4)
                while len(header) < 4:
                    header += socket.recv(4 - len(header))
                size = _struct.unpack('I', header)[0]

                message = socket.recv(size)
                while len(message) < size:
                    message += socket.recv(size)
                self.dispatch_event('on_receive', self, message)
            except (ConnectionError, OSError):
                self.close()
                break

        assert _debug_net("Exiting _recv thread")

    def on_receive(self, connection, message):
        """Event for received messages."""

    def on_disconnect(self, connection):
        """Event for disconnection. """

    def __repr__(self):
        return f"Client(address={self._address}, port={self._port})"


Client.register_event_type('on_receive')
Client.register_event_type('on_disconnect')


class ClientConnection(_EventDispatcher):

    def __init__(self, reader, writer):
        self._reader = reader
        self._writer = writer
        self._closed = False
        self._loop = _asyncio.get_event_loop()
        _asyncio.run_coroutine_threadsafe(self._recv(), self._loop)

    def close(self):
        if not self._closed:
            self._writer.transport.close()
            self._closed = True
            self.dispatch_event('on_disconnect', self)

    async def _recv(self):
        while not self._closed:
            try:
                header = await self._reader.readexactly(4)
                size = _struct.unpack('I', header)[0]
                message = await self._reader.readexactly(size)
                self._loop.call_soon(self.dispatch_event, 'on_receive', self, message)

            except _asyncio.IncompleteReadError:
                self.close()
                break

    async def _send(self, message):
        try:
            packet = _struct.pack('I', len(message)) + message
            self._writer.write(packet)
            await self._writer.drain()
        except ConnectionResetError:
            self.close()

    def send(self, message):
        # Synchrounously send a message in an async coroutine.
        if self._writer.transport is None or self._writer.transport.is_closing():
            self.close()
            return
        _future = _asyncio.run_coroutine_threadsafe(self._send(message), self._loop)

    def on_receive(self, connection, message):
        """Event for received messages."""

    def on_disconnect(self, connection):
        """Event for disconnection. """

    def __del__(self):
        assert _debug_net(f"Connection garbage collected: {self}")

    def __repr__(self):
        return f"{self.__class__.__name__}({id(self)})"


ClientConnection.register_event_type('on_receive')
ClientConnection.register_event_type('on_disconnect')


class Server(_EventDispatcher):

    def __init__(self, address, port):
        self._address = address
        self._port = port

        self._server = None

        self._thread = _threading.Thread(target=self._run, daemon=True)
        self._thread.start()

        blurb = f"Server listening on {address}:{port}"
        assert _debug_net(f"{'-' * len(blurb)}\n{blurb}\n{'-' * len(blurb)}")


    async def handle_connection(self, reader, writer):
        connection = ClientConnection(reader, writer)
        self.dispatch_event('on_connection', connection)

    async def _start_server(self):
        self._server = await _asyncio.start_server(self.handle_connection, self._address, self._port)
        async with self._server:
            await self._server.serve_forever()

    def _run(self):
        try:
            _asyncio.run(self._start_server())
        except KeyboardInterrupt:
            self._server.close()

    def on_connection(self, connection):
        """Event for new Client connections."""
        assert _debug_net(f"Connected <--- {connection}")
        connection.set_handler('on_disconnect', self.on_disconnect)

    def on_disconnect(self, connection):
        """Event for disconnected Clients."""
        assert _debug_net(f"Disconnected ---> {connection}")


Server.register_event_type('on_connection')
Server.register_event_type('on_disconnect')



"""OpenGL Framebuffer abstractions.

This module provides classes for working with Framebuffers & Renderbuffers
and their attachments. Attachments can be pyglet Texture objects, which allows
easily accessing their data, saving to disk, etc. Renderbuffers can be used
if you don't need to access their data at a later time. For example::

    # Create two objects to use as attachments for our Framebuffer.
    color_buffer = pyglet.image.Texture.create(width, height, min_filter=GL_NEAREST, mag_filter=GL_NEAREST)
    depth_buffer = pyglet.image.buffer.Renderbuffer(width, height, GL_DEPTH_COMPONENT)

    # Create a framebuffer object, and attach the two buffers:
    framebuffer = pyglet.image.Framebuffer()
    framebuffer.attach_texture(color_buffer, attachment=GL_COLOR_ATTACHMENT0)
    framebuffer.attach_renderbuffer(depth_buffer, attachment=GL_DEPTH_ATTACHMENT)

    # Bind the Framebuffer, which sets it as the active render target:
    framebuffer.bind()

See the OpenGL documentation for more information on valid attachment types and targets.
"""
from __future__ import annotations

import pyglet

from pyglet.gl import GLint, GLuint, glBindFramebuffer, glBindRenderbuffer, glCheckFramebufferStatus, glClear
from pyglet.gl import glDeleteFramebuffers, glDeleteRenderbuffers, glFramebufferRenderbuffer, glFramebufferTexture
from pyglet.gl import glFramebufferTextureLayer, glGenFramebuffers, glGenRenderbuffers, glGetIntegerv
from pyglet.gl import glRenderbufferStorage, glRenderbufferStorageMultisample
from pyglet.gl import GL_COLOR_ATTACHMENT0, GL_MAX_COLOR_ATTACHMENTS, GL_RENDERBUFFER
from pyglet.gl import GL_FRAMEBUFFER, GL_FRAMEBUFFER_COMPLETE, GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT
from pyglet.gl import GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT, GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER
from pyglet.gl import GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT, GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT
from pyglet.gl import GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER, GL_FRAMEBUFFER_UNSUPPORTED

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyglet.image import Texture


def get_max_color_attachments() -> int:
    """Get the maximum allow Framebuffer Color attachments."""
    number = GLint()
    glGetIntegerv(GL_MAX_COLOR_ATTACHMENTS, number)
    return number.value


class Renderbuffer:
    """OpenGL Renderbuffer Object."""

    def __init__(self, width: int, height: int, internal_format: int, samples: int = 1) -> None:
        """Create a RenderBuffer instance."""
        self._context = pyglet.gl.current_context
        self._id = GLuint()
        self._width = width
        self._height = height
        self._internal_format = internal_format

        glGenRenderbuffers(1, self._id)
        glBindRenderbuffer(GL_RENDERBUFFER, self._id)

        if samples > 1:
            glRenderbufferStorageMultisample(GL_RENDERBUFFER, samples, internal_format, width, height)
        else:
            glRenderbufferStorage(GL_RENDERBUFFER, internal_format, width, height)

        glBindRenderbuffer(GL_RENDERBUFFER, 0)

    @property
    def id(self) -> int:
        return self._id.value

    @property
    def width(self) -> int:
        return self._width

    @property
    def height(self) -> int:
        return self._height

    def bind(self) -> None:
        glBindRenderbuffer(GL_RENDERBUFFER, self._id)

    @staticmethod
    def unbind() -> None:
        glBindRenderbuffer(GL_RENDERBUFFER, 0)

    def delete(self) -> None:
        glDeleteRenderbuffers(1, self._id)
        self._id = None

    def __del__(self) -> None:
        if self._id is not None:
            try:
                self._context.delete_renderbuffer(self._id.value)
                self._id = None
            except (AttributeError, ImportError):
                pass  # Interpreter is shutting down

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self._id.value})"


class Framebuffer:
    """OpenGL Framebuffer Object.

    .. versionadded:: 2.0
    """
    def __init__(self, target: int = GL_FRAMEBUFFER) -> None:
        """Create a Framebuffer Instance."""
        self._context = pyglet.gl.current_context
        self._id = GLuint()
        glGenFramebuffers(1, self._id)
        self._attachment_types = 0
        self._width = 0
        self._height = 0
        self.target = target

    @property
    def id(self) -> int:
        """The Framebuffer id."""
        return self._id.value

    @property
    def width(self) -> int:
        """The width of the widest attachment."""
        return self._width

    @property
    def height(self) -> int:
        """The height of the tallest attachment."""
        return self._height

    def bind(self) -> None:
        """Bind the Framebuffer.

        This ctivates it as the current drawing target.
        """
        glBindFramebuffer(self.target, self._id)

    def unbind(self) -> None:
        """Unbind the Framebuffer.

        Unbind should be called to prevent further rendering
        to the framebuffer, or if you wish to access data
        from its Texture attachments.
        """
        glBindFramebuffer(self.target, 0)

    def clear(self) -> None:
        """Clear the attachments."""
        if self._attachment_types:
            self.bind()
            glClear(self._attachment_types)
            self.unbind()

    def delete(self) -> None:
        """Explicitly delete the Framebuffer."""
        glDeleteFramebuffers(1, self._id)
        self._id = None

    def __del__(self) -> None:
        if self._id is not None:
            try:
                self._context.delete_framebuffer(self._id.value)
                self._id = None
            except (AttributeError, ImportError):
                pass  # Interpreter is shutting down

    @property
    def is_complete(self) -> bool:
        """True if the framebuffer is 'complete', else False."""
        return glCheckFramebufferStatus(GL_FRAMEBUFFER) == GL_FRAMEBUFFER_COMPLETE

    @staticmethod
    def get_status() -> str:
        """Get the current Framebuffer status, as a string.

        If ``Framebuffer.is_complete`` is ``False``, this method
        can be used for more information. It will return a
        string with the OpenGL reported status.
        """
        states = {GL_FRAMEBUFFER_UNSUPPORTED: "Framebuffer unsupported. Try another format.",
                  GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT: "Framebuffer incomplete attachment.",
                  GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT: "Framebuffer missing attachment.",
                  GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT: "Framebuffer unsupported dimension.",
                  GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT: "Framebuffer incomplete formats.",
                  GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER: "Framebuffer incomplete draw buffer.",
                  GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER: "Framebuffer incomplete read buffer.",
                  GL_FRAMEBUFFER_COMPLETE: "Framebuffer is complete."}

        gl_status = glCheckFramebufferStatus(GL_FRAMEBUFFER)

        return states.get(gl_status, "Unknown error")

    def attach_texture(self, texture: Texture,
                       target: int = GL_FRAMEBUFFER, attachment: int = GL_COLOR_ATTACHMENT0) -> None:
        """Attach a Texture to the Framebuffer.

        Args:
            texture:
                Specifies the texture object to attach to the framebuffer attachment
                point named by attachment.
            target:
                Specifies the framebuffer target. target must be GL_DRAW_FRAMEBUFFER,
                GL_READ_FRAMEBUFFER, or GL_FRAMEBUFFER. GL_FRAMEBUFFER is equivalent
                to GL_DRAW_FRAMEBUFFER.
            attachment:
                Specifies the attachment point of the framebuffer. attachment must be
                GL_COLOR_ATTACHMENTi, GL_DEPTH_ATTACHMENT, GL_STENCIL_ATTACHMENT or
                GL_DEPTH_STENCIL_ATTACHMENT.
        """
        self.bind()
        glFramebufferTexture(target, attachment, texture.id, texture.level)
        self._attachment_types |= attachment
        self._width = max(texture.width, self._width)
        self._height = max(texture.height, self._height)
        self.unbind()

    def attach_texture_layer(self, texture: Texture, layer: int, level: int,
                             target: int = GL_FRAMEBUFFER, attachment: int = GL_COLOR_ATTACHMENT0) -> None:
        """Attach a Texture layer to the Framebuffer.

        Args:
            texture:
                Specifies the texture object to attach to the framebuffer attachment
                point named by attachment.
            layer:
                Specifies the layer of texture to attach.
            level:
                Specifies the mipmap level of texture to attach.
            target:
                Specifies the framebuffer target. target must be GL_DRAW_FRAMEBUFFER,
                GL_READ_FRAMEBUFFER, or GL_FRAMEBUFFER. GL_FRAMEBUFFER is equivalent
                to GL_DRAW_FRAMEBUFFER.
            attachment:
                Specifies the attachment point of the framebuffer. attachment must be
                GL_COLOR_ATTACHMENTi, GL_DEPTH_ATTACHMENT, GL_STENCIL_ATTACHMENT or
                GL_DEPTH_STENCIL_ATTACHMENT.
        """
        self.bind()
        glFramebufferTextureLayer(target, attachment, texture.id, level, layer)
        self._attachment_types |= attachment
        self._width = max(texture.width, self._width)
        self._height = max(texture.height, self._height)
        self.unbind()

    def attach_renderbuffer(self, renderbuffer: Renderbuffer,
                            target: int = GL_FRAMEBUFFER, attachment: int = GL_COLOR_ATTACHMENT0) -> None:
        """Attach a Renderbuffer to the Framebuffer.

        Args:
            renderbuffer:
                Specifies the Renderbuffer to attach to the framebuffer attachment
                point named by attachment.
            target:
                Specifies the framebuffer target. target must be GL_DRAW_FRAMEBUFFER,
                GL_READ_FRAMEBUFFER, or GL_FRAMEBUFFER. GL_FRAMEBUFFER is equivalent
                to GL_DRAW_FRAMEBUFFER.
            attachment:
                Specifies the attachment point of the framebuffer. attachment must be
                GL_COLOR_ATTACHMENTi, GL_DEPTH_ATTACHMENT, GL_STENCIL_ATTACHMENT or
                GL_DEPTH_STENCIL_ATTACHMENT.
        """
        self.bind()
        glFramebufferRenderbuffer(target, attachment, GL_RENDERBUFFER, renderbuffer.id)
        self._attachment_types |= attachment
        self._width = max(renderbuffer.width, self._width)
        self._height = max(renderbuffer.height, self._height)
        self.unbind()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self._id.value})"



#!/usr/bin/env python
"""Display an image.

Usage::

    display.py <filename>

A checkerboard background is visible behind any transparent areas.
"""


import sys

import pyglet
from pyglet.gl import (
    glEnable,
    glBlendFunc,
    GL_BLEND,
    GL_ONE_MINUS_SRC_ALPHA,
    GL_SRC_ALPHA,
)

window = pyglet.window.Window(visible=False, resizable=True)


@window.event
def on_draw():
    background.blit_tiled(0, 0, 0, window.width, window.height)
    img.blit(window.width // 2, window.height // 2, 0)


if __name__ == '__main__':
    if len(sys.argv) != 2:
        print(__doc__)
        sys.exit(1)

    filename = sys.argv[1]

    img = pyglet.image.load(filename)
    img.anchor_x = img.width // 2
    img.anchor_y = img.height // 2

    checks = pyglet.image.create(32, 32, pyglet.image.CheckerImagePattern())
    background = pyglet.image.TileableTexture.create_for_image(checks)

    # Enable alpha blending, required for image.blit.
    glEnable(GL_BLEND)
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

    window.width = img.width
    window.height = img.height
    window.set_visible()

    pyglet.app.run()



"""
Usage

    playmany.py

Uses media_player to play a sequence of samples and record debug info

A configuration must be active, see command configure.py
If the active configuration has disallowed dbg overwrites it will do nothing.

If a playlist was provided at session creation, then only the samples in the
playlist will be played, otherwise all files in samples_dir.

"""

import os
import subprocess
import sys

import fs
import mpexceptions


def main():
    try:
        pathserv = fs.get_path_info_for_active_session()
    except mpexceptions.ExceptionUndefinedSamplesDir:
        print("The env var 'pyglet_mp_samples_dir' is not defined.")
        return 1
    except mpexceptions.ExceptionNoSessionIsActive:
        print("*** Error, no session active.")
        return 1

    try:
        play_many(pathserv, timeout=120)
    except mpexceptions.ExceptionAttemptToBreakRawDataProtection:
        print("*** Error, attempt to overwrite raw data when protect_raw_data is True.")
        return 1

    return 0


def play_many(pathserv, timeout=120):
    """plays the samples in the session playlist for the current active session
       timeout: max time allowed to play a sample, default is 120 seconds
    """
    conf = fs.get_session_configuration(pathserv)
    if conf["dev_debug"]:
        pass
    else:
        if conf["protect_raw_data"]:
            raise mpexceptions.ExceptionAttemptToBreakRawDataProtection()

    playlist_gen = pathserv.session_playlist_generator()
    core_play_many(pathserv, playlist_gen, timeout=timeout)


def core_play_many(pathserv, playlist_gen, timeout=120):
    for sample, filename in playlist_gen:
        dbg_file = pathserv.dbg_filename(sample)

        print("playmany playing:", filename)

        cmdline = [os.path.join(fs.get_media_player_path(), "media_player.py"),
                   "--debug",
                   "--outfile=" + dbg_file,
                   filename]
        killed, returncode = cmd__py3(cmdline, timeout=timeout)
        if killed:
            print("WARNING: killed by timeout, file: %s" % filename)


def cmd__py3(cmdline, bufsize=-1, cwd=None, timeout=60):
    """runs a .py script as a subprocess with the same python as the caller

       cmdline: list [<scriptname>, arg1, ...]
       timeout: time in seconds; subprocess will be killed if it is still running
                at that time.
    """
    # use the same python as the caller to run the script
    cmdline.insert(0, "-u")
    cmdline.insert(0, sys.executable)

    p = subprocess.Popen(
        cmdline,
        bufsize = bufsize,
        shell   = False,
        stdout  = subprocess.PIPE,
        stderr  = subprocess.PIPE,
        cwd     = cwd
    )
    killed = True
    try:
        out, err = p.communicate(timeout=timeout)
        killed = False
    except subprocess.TimeoutExpired:
        p.kill()
        out, err = p.communicate()
##    print("out:", out)
##    print("err:", err)

    returncode = p.returncode

    return killed, returncode


def sysargs_to_mainargs():
    """builds main args from sys.argv"""
    if len(sys.argv) > 1 and sys.argv[1].startswith("--help"):
        print(__doc__)
        sys.exit(1)

if __name__ == "__main__":
    sysargs_to_mainargs()
    main()



"""Generate a Python ctypes wrapper file for a header file.

Usage example::
    wrap.py -lGL -oGL.py /usr/include/GL/gl.h

    >>> from GL import *

"""
from __future__ import annotations
import sys
import textwrap

from ctypesparser import *


class CtypesWrapper(CtypesParser, CtypesTypeVisitor):
    file=None
    def begin_output(self, output_file, library, link_modules=(), 
                     emit_filenames=(), all_headers=False):
        self.library = library
        self.file = output_file
        self.all_names = []
        self.known_types = {}
        self.structs = set()
        self.enums = set()
        self.emit_filenames = emit_filenames
        self.all_headers = all_headers

        self.linked_symbols = {}
        for name in link_modules:
            module = __import__(name, globals(), locals(), ['foo'])
            for symbol in dir(module):
                if symbol not in self.linked_symbols:
                    self.linked_symbols[symbol] = '%s.%s' % (name, symbol)
        self.link_modules = link_modules

        self.print_preamble()
        self.print_link_modules_imports()

    def wrap(self, filename, source=None):
        assert self.file, 'Call begin_output first'
        self.parse(filename, source)

    def end_output(self):
        self.print_epilogue()
        self.file = None

    def does_emit(self, symbol, filename):
        return self.all_headers or filename in self.emit_filenames

    def print_preamble(self):
        import textwrap
        print(textwrap.dedent("""
            '''Wrapper for {library}

            Generated with:
            {argv}

            Do not modify this file.
            '''
            import ctypes
            from ctypes import *

            import pyglet.lib

            _lib = pyglet.lib.load_library({library!r})

            _int_types = (c_int16, c_int32)
            if hasattr(ctypes, 'c_int64'):
                # Some builds of ctypes apparently do not have c_int64
                # defined; it's a pretty good bet that these builds do not
                # have 64-bit pointers.
                _int_types += (ctypes.c_int64,)
            for t in _int_types:
                if sizeof(t) == sizeof(c_size_t):
                    c_ptrdiff_t = t

            class c_void(Structure):
                # c_void_p is a buggy return type, converting to int, so
                # POINTER(None) == c_void_p is actually written as
                # POINTER(c_void), so it can be treated as a real pointer.
                _fields_ = [('dummy', c_int)]

        """.format(
            library=self.library,
            argv=' '.join(sys.argv),
        )).lstrip(), file=self.file)

    def print_link_modules_imports(self) -> None:
        for name in self.link_modules:
            print(f'import {name}', file=self.file)
        print(file=self.file)

    def print_epilogue(self) -> None:
        print(file=self.file)
        print('\n'.join(textwrap.wrap(
            '__all__ = [{}]'.format(', '.join([repr(n) for n in self.all_names])),
            width=78,
            break_long_words=False)), file=self.file)

    def handle_ctypes_constant(self, name, value, filename, lineno) -> None:
        if self.does_emit(name, filename):
            print(f'{name} = {value!r}', end=' ', file=self.file)
            print('\t# %s:%d' % (filename, lineno), file=self.file)
            self.all_names.append(name)

    def handle_ctypes_type_definition(self, name, ctype, filename, lineno) -> None:
        if self.does_emit(name, filename):
            self.all_names.append(name)
            if name in self.linked_symbols:
                print(f'{name} = {self.linked_symbols[name]}', file=self.file)
            else:
                ctype.visit(self)
                self.emit_type(ctype)
                print(f'{name} = {ctype!s}', end=' ', file=self.file)
                print('\t# %s:%d' % (filename, lineno), file=self.file)
        else:
            self.known_types[name] = (ctype, filename, lineno)

    def emit_type(self, t) -> None:
        t.visit(self)
        for s in t.get_required_type_names():
            if s in self.known_types:
                if s in self.linked_symbols:
                    print(f'{s} = {self.linked_symbols[s]}', file=self.file)
                else:
                    s_ctype, s_filename, s_lineno = self.known_types[s]
                    s_ctype.visit(self)

                    self.emit_type(s_ctype)
                    print(f'{s} = {s_ctype!s}', end=' ', file=self.file)
                    print('\t# %s:%d' % (s_filename, s_lineno), file=self.file)
                del self.known_types[s]

    def visit_struct(self, struct) -> None:
        if struct.tag in self.structs:
            return
        self.structs.add(struct.tag)

        base = {True: 'Union', False: 'Structure'}[struct.is_union]
        print(f'class struct_{struct.tag}({base}):', file=self.file)
        print('    __slots__ = [', file=self.file)
        if not struct.opaque:
            for m in struct.members:
                print(f"        '{m[0]}',", file=self.file)
        print('    ]', file=self.file)

        # Set fields after completing class, so incomplete structs can be
        # referenced within struct.
        for _name, typ in struct.members:
            self.emit_type(typ)

        print(f'struct_{struct.tag}._fields_ = [', file=self.file)
        if struct.opaque:
            print("    ('_opaque_struct', c_int)", file=self.file)
            self.structs.remove(struct.tag)
        else:
            for m in struct.members:
                print(f"    ('{m[0]}', {m[1]}),", file=self.file)
        print(']', file=self.file)
        print(file=self.file)

    def visit_enum(self, enum) -> None:
        if enum.tag in self.enums:
            return
        self.enums.add(enum.tag)

        print(f'enum_{enum.tag} = c_int', file=self.file)
        for name, value in enum.enumerators:
            self.all_names.append(name)
            print('%s = %d' % (name, value), file=self.file)

    def handle_ctypes_function(self, name, restype, argtypes, filename, lineno) -> None:
        if self.does_emit(name, filename):
            # Also emit any types this func requires that haven't yet been
            # written.
            self.emit_type(restype)
            for a in argtypes:
                self.emit_type(a)

            self.all_names.append(name)
            print('# %s:%d' % (filename, lineno), file=self.file)
            print(f'{name} = _lib.{name}', file=self.file)
            print(f'{name}.restype = {restype!s}', file=self.file)
            print('{}.argtypes = [{}]'.format(name, ', '.join([str(a) for a in argtypes])), file=self.file)
            print(file=self.file)

    def handle_ctypes_variable(self, name, ctype, filename, lineno) -> None:
        # This doesn't work.  (but I(shenjackyuanjie) still remove the '%' content
        # self.all_names.append(name)
        # print >> self.file, f'{name} = {str(ctype)}.indll(_lib, {name!r})'
        pass

def main(*argv) -> None:
    import optparse
    import os.path
    import sys

    usage = 'usage: %prog [options] <header.h>'
    op = optparse.OptionParser(usage=usage)
    op.add_option('-o', '--output', dest='output',
                  help='write wrapper to FILE', metavar='FILE')
    op.add_option('-l', '--library', dest='library',
                  help='link to LIBRARY', metavar='LIBRARY')
    op.add_option('-D', '--define', dest='defines', default=[],
                  help='define token NAME=VALUE', action='append')
    op.add_option('-i', '--include-file', action='append', dest='include_files',
                  help='assume FILE is iincluded', metavar='FILE',
                  default=[])
    op.add_option('-I', '--include-dir', action='append', dest='include_dirs',
                  help='add DIR to include search path', metavar='DIR',
                  default=[])
    op.add_option('-m', '--link-module', action='append', dest='link_modules',
                  help='use symbols from MODULE', metavar='MODULE',
                  default=[])
    op.add_option('-a', '--all-headers', action='store_true',
                  dest='all_headers',
                  help='include symbols from all headers', default=False)

    (options, args) = op.parse_args(list(argv[1:]))
    if len(args) < 1:
        print('No header files specified.', file=sys.stderr) 
        sys.exit(1)
    headers = args

    if options.library is None:
        options.library = os.path.splitext(headers[0])[0]
    if options.output is None:
        options.output = f'{options.library}.py'

    wrapper = CtypesWrapper()
    wrapper.begin_output(open(options.output, 'w'),
                         library=options.library,
                         emit_filenames=headers,
                         link_modules=options.link_modules,
                         all_headers=options.all_headers)
    wrapper.preprocessor_parser.include_path += options.include_dirs
    for define in options.defines:
        name, value = define.split('=')
        wrapper.preprocessor_parser.define(name, value)
    for file in options.include_files:
        wrapper.wrap(file)
    for header in headers:
        wrapper.wrap(header)
    wrapper.end_output()

    print(f'Wrapped to {options.output}')

if __name__ == '__main__':
    main(*sys.argv)





def pytest_addoption(parser):
    """Add the special options for interactive tests."""
    parser.addoption('--non-interactive',
                     action='store_true',
                     help='[Interactive tests only] Do not use interactive prompts. '
                          'Skip tests that cannot validate or run without.',
                     )
    parser.addoption('--sanity',
                     action='store_true',
                     help='[Interactive tests only] Do not use interactive prompts. '
                          'Only skips tests that cannot finish without user intervention.',
                     )


# Import shared fixtures
from .base.data import test_data
from .base.performance import performance




import warnings
from ctypes import c_ulong, c_int, POINTER, Structure, c_char_p, c_uint, c_ushort

import pyglet
from pyglet.libs.x11.xlib import (
    Time,
    Window,
    Display,
    XCloseDisplay,
    XDefaultRootWindow,
    XOpenDisplay,
)

lib = None
try:
    lib = pyglet.lib.load_library("Xrandr")
except ImportError:
    if pyglet.options.debug_lib:
        warnings.warn("Xrandr could not be loaded.")
    raise ImportError

RRCrtc      = c_ulong        # typedef XID
RROutput    = c_ulong        # typedef XID
RRMode      = c_ulong        # typedef XID
Connection  = c_ushort       # Connection in Xrandr.h
SubpixelOrder = c_ushort     # SubpixelOrder in Xrandr.h
XRRModeFlags = c_ulong

class XRRModeInfo(Structure):
    _fields_ = [
        ("id",          RRMode),     # mode ID
        ("width",       c_uint),   # horizontal resolution
        ("height",      c_uint),   # vertical resolution
        ("dotClock",    c_ulong),    # pixel clock (in Hz)
        ("hSyncStart",  c_uint),
        ("hSyncEnd",    c_uint),
        ("hTotal",      c_uint),
        ("hSkew",       c_uint),
        ("vSyncStart",  c_uint),
        ("vSyncEnd",    c_uint),
        ("vTotal",      c_uint),
        ("name",        c_char_p),
        ("nameLength",  c_uint),
        ("modeFlags",   XRRModeFlags),
    ]


class XRRScreenResources(Structure):
    _fields_ = [
        ("timestamp",        Time),
        ("configTimestamp",  Time),
        ("ncrtc",            c_int),
        ("crtcs",            POINTER(RRCrtc)),
        ("noutput",          c_int),
        ("outputs",          POINTER(RROutput)),
        ("nmode",            c_int),
        ("modes",            POINTER(XRRModeInfo)),
    ]


class XRROutputInfo(Structure):
    _fields_ = [
        ("timestamp",      Time),
        ("crtc",           RRCrtc),
        ("name",           c_char_p),
        ("nameLen",        c_int),
        ("mm_width",       c_ulong),
        ("mm_height",      c_ulong),
        ("connection",     Connection),
        ("subpixel_order", SubpixelOrder),
        ("ncrtc",          c_int),
        ("crtcs",          POINTER(RRCrtc)),
        ("nclone",         c_int),
        ("clones",         POINTER(RROutput)),
        ("nmode",          c_int),
        ("npreferred",     c_int),
        ("modes",          POINTER(RRMode)),
    ]

class XRRCrtcInfo(Structure):
    _fields_ = [
        ("timestamp",  Time),
        ("x",          c_int),
        ("y",          c_int),
        ("width",      c_uint),
        ("height",     c_uint),
        ("mode",       RRMode),
        ("rotation",   c_int),
        ("noutput",    c_int),
        ("outputs",    POINTER(RROutput)),
        ("rotations",  c_ushort),
        ("npossible",  c_int),
        ("possible",   POINTER(RROutput)),
    ]


if lib:
    XRRQueryVersion = lib.XRRQueryVersion
    XRRQueryVersion.argtypes = [POINTER(Display), POINTER(c_int), POINTER(c_int)]
    XRRQueryVersion.restype    = c_int

    XRRGetScreenResources = lib.XRRGetScreenResources
    XRRGetScreenResources.argtypes = [POINTER(Display), Window]
    XRRGetScreenResources.restype = POINTER(XRRScreenResources)

    XRRGetOutputPrimary = lib.XRRGetOutputPrimary
    XRRGetOutputPrimary.argtypes = [POINTER(Display), Window]
    XRRGetOutputPrimary.restype = RROutput

    XRRGetScreenResourcesCurrent = lib.XRRGetScreenResourcesCurrent
    XRRGetScreenResourcesCurrent.argtypes = [POINTER(Display), Window]
    XRRGetScreenResourcesCurrent.restype = POINTER(XRRScreenResources)

    XRRFreeScreenResources = lib.XRRFreeScreenResources
    XRRFreeScreenResources.argtypes = [POINTER(XRRScreenResources)]
    XRRFreeScreenResources.restype = None

    XRRGetOutputInfo = lib.XRRGetOutputInfo
    XRRGetOutputInfo.argtypes = [POINTER(Display), POINTER(XRRScreenResources), RROutput]
    XRRGetOutputInfo.restype  = POINTER(XRROutputInfo)

    XRRFreeOutputInfo = lib.XRRFreeOutputInfo
    XRRFreeOutputInfo.argtypes = [POINTER(XRROutputInfo)]
    XRRFreeOutputInfo.restype = None

    XRRGetCrtcInfo = lib.XRRGetCrtcInfo
    XRRGetCrtcInfo.argtypes = [POINTER(Display), POINTER(XRRScreenResources), RRCrtc]
    XRRGetCrtcInfo.restype = POINTER(XRRCrtcInfo)

    XRRSetCrtcConfig = lib.XRRSetCrtcConfig
    XRRSetCrtcConfig.argtypes = argtypes = [POINTER(Display), POINTER(XRRScreenResources), RRCrtc, Time, c_int, c_int, RRMode, c_int,  POINTER(RROutput), c_int]
    XRRSetCrtcConfig.restype = c_int

    XRRFreeCrtcInfo = lib.XRRFreeCrtcInfo
    XRRFreeCrtcInfo.argtypes = [POINTER(XRRCrtcInfo)]
    XRRFreeCrtcInfo.restype    = None

def list_connected_outputs():
    dpy = XOpenDisplay(None)
    if not dpy:
        raise RuntimeError("Cannot open DISPLAY")

    root = XDefaultRootWindow(dpy)
    res_p = XRRGetScreenResources(dpy, root)
    if not res_p:
        XCloseDisplay(dpy)
        raise RuntimeError("Failed to get screen resources")

    res = res_p.contents
    outputs = []
    for i in range(res.noutput):
        out_id = res.outputs[i]
        info_p = XRRGetOutputInfo(dpy, res_p, out_id)
        if info_p:
            info = info_p.contents
            if info.connection == 0:  # Connected.
                outputs.append(info.name.decode())
            XRRFreeOutputInfo(info_p)

    XRRFreeScreenResources(res_p)
    XCloseDisplay(dpy)
    return outputs

if __name__ == "__main__":
    for name in list_connected_outputs():
        print(name)